import os

from torchvision import transforms
from torchvision.datasets import CIFAR100

from nebula.core.datasets.nebuladataset import NebulaDataset


class CIFAR100Dataset(NebulaDataset):
    def __init__(
        self,
        num_classes=100,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__(
            num_classes=num_classes,
            partition_id=partition_id,
            partitions_number=partitions_number,
            batch_size=batch_size,
            num_workers=num_workers,
            iid=iid,
            partition=partition,
            partition_parameter=partition_parameter,
            seed=seed,
            config=config,
        )

    def initialize_dataset(self):
        # Load CIFAR100 train dataset
        if self.train_set is None:
            self.train_set = self.load_cifar100_dataset(train=True)
        if self.test_set is None:
            self.test_set = self.load_cifar100_dataset(train=False)

        # All nodes have the same test set (indices are the same for all nodes)
        self.test_indices_map = list(range(len(self.test_set)))

        # Depending on the iid flag, generate a non-iid or iid map of the train set
        if self.iid:
            self.train_indices_map = self.generate_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_iid_map(self.test_set, self.partition, self.partition_parameter)
        else:
            self.train_indices_map = self.generate_non_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_non_iid_map(
                self.test_set, self.partition, self.partition_parameter
            )

        print(f"Length of train indices map: {len(self.train_indices_map)}")
        print(f"Lenght of test indices map (global): {len(self.test_indices_map)}")
        print(f"Length of test indices map (local): {len(self.local_test_indices_map)}")

    def load_cifar100_dataset(self, train=True):
        mean = (0.4914, 0.4822, 0.4465)
        std = (0.2471, 0.2435, 0.2616)
        apply_transforms = transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize(mean, std, inplace=True),
        ])
        return CIFAR100(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "data"),
            train=train,
            download=True,
            transform=apply_transforms,
        )

    def generate_non_iid_map(self, dataset, partition="dirichlet", partition_parameter=0.5):
        if partition == "dirichlet":
            partitions_map = self.dirichlet_partition(dataset, alpha=partition_parameter)
        elif partition == "percent":
            partitions_map = self.percentage_partition(dataset, percentage=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for Non-IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]

    def generate_iid_map(self, dataset, partition="balancediid", partition_parameter=2):
        if partition == "balancediid":
            partitions_map = self.balanced_iid_partition(dataset)
        elif partition == "unbalancediid":
            partitions_map = self.unbalanced_iid_partition(dataset, imbalance_factor=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]



import os
import random
import sys
import time
import warnings

import numpy as np
import torch

torch.multiprocessing.set_start_method("spawn", force=True)

# Ignore CryptographyDeprecationWarning (datatime issues with cryptography library)
from cryptography.utils import CryptographyDeprecationWarning

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
import logging

from nebula.config.config import Config
from nebula.core.datasets.cifar10.cifar10 import CIFAR10Dataset
from nebula.core.datasets.cifar100.cifar100 import CIFAR100Dataset
from nebula.core.datasets.datamodule import DataModule
from nebula.core.datasets.emnist.emnist import EMNISTDataset
from nebula.core.datasets.fashionmnist.fashionmnist import FashionMNISTDataset
from nebula.core.datasets.kitsun.kitsun import KITSUNDataset
from nebula.core.datasets.militarysar.militarysar import MilitarySARDataset
from nebula.core.datasets.mnist.mnist import MNISTDataset
from nebula.core.datasets.syscall.syscall import SYSCALLDataset
from nebula.core.engine import AggregatorNode, IdleNode, MaliciousNode, ServerNode, TrainerNode
from nebula.core.models.cifar10.cnn import CIFAR10ModelCNN
from nebula.core.models.cifar10.cnnV2 import CIFAR10ModelCNN_V2
from nebula.core.models.cifar10.cnnV3 import CIFAR10ModelCNN_V3
from nebula.core.models.cifar10.dualagg import DualAggModel
from nebula.core.models.cifar10.fastermobilenet import FasterMobileNet
from nebula.core.models.cifar10.resnet import CIFAR10ModelResNet
from nebula.core.models.cifar10.simplemobilenet import SimpleMobileNetV1
from nebula.core.models.cifar100.cnn import CIFAR100ModelCNN
from nebula.core.models.emnist.cnn import EMNISTModelCNN
from nebula.core.models.emnist.mlp import EMNISTModelMLP
from nebula.core.models.fashionmnist.cnn import FashionMNISTModelCNN
from nebula.core.models.fashionmnist.mlp import FashionMNISTModelMLP
from nebula.core.models.kitsun.mlp import KitsunModelMLP
from nebula.core.models.militarysar.cnn import MilitarySARModelCNN
from nebula.core.models.mnist.cnn import MNISTModelCNN
from nebula.core.models.mnist.mlp import MNISTModelMLP
from nebula.core.models.syscall.autoencoder import SyscallModelAutoencoder
from nebula.core.models.syscall.mlp import SyscallModelMLP
from nebula.core.models.syscall.svm import SyscallModelSGDOneClassSVM
from nebula.core.role import Role
from nebula.core.training.lightning import Lightning
from nebula.core.training.siamese import Siamese

# os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
# os.environ["TORCH_LOGS"] = "+dynamo"
# os.environ["TORCHDYNAMO_VERBOSE"] = "1"


async def main(config):
    n_nodes = config.participant["scenario_args"]["n_nodes"]
    model_name = config.participant["model_args"]["model"]
    idx = config.participant["device_args"]["idx"]

    additional_node_status = config.participant["mobility_args"]["additional_node"]["status"]
    additional_node_round = config.participant["mobility_args"]["additional_node"]["round_start"]

    attacks = config.participant["adversarial_args"]["attacks"]
    poisoned_persent = config.participant["adversarial_args"]["poisoned_sample_percent"]
    poisoned_ratio = config.participant["adversarial_args"]["poisoned_ratio"]
    targeted = str(config.participant["adversarial_args"]["targeted"])
    target_label = config.participant["adversarial_args"]["target_label"]
    target_changed_label = config.participant["adversarial_args"]["target_changed_label"]
    noise_type = config.participant["adversarial_args"]["noise_type"]
    iid = config.participant["data_args"]["iid"]
    partition_selection = config.participant["data_args"]["partition_selection"]
    partition_parameter = np.array(config.participant["data_args"]["partition_parameter"], dtype=np.float64)
    label_flipping = False
    data_poisoning = False
    model_poisoning = False
    if attacks == "Label Flipping":
        label_flipping = True
        poisoned_ratio = 0
        if targeted == "true" or targeted == "True":
            targeted = True
        else:
            targeted = False
    elif attacks == "Sample Poisoning":
        data_poisoning = True
        if targeted == "true" or targeted == "True":
            targeted = True
        else:
            targeted = False
    elif attacks == "Model Poisoning":
        model_poisoning = True
    else:
        label_flipping = False
        data_poisoning = False
        targeted = False
        poisoned_persent = 0
        poisoned_ratio = 0

    # Adjust the total number of nodes and the index of the current node for CFL, as it doesn't require a specific partition for the server (not used for training)
    if config.participant["scenario_args"]["federation"] == "CFL":
        n_nodes -= 1
        if idx > 0:
            idx -= 1

    dataset = None
    dataset_str = config.participant["data_args"]["dataset"]
    num_workers = config.participant["data_args"]["num_workers"]
    model = None
    if dataset_str == "MNIST":
        dataset = MNISTDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "MLP":
            model = MNISTModelMLP()
        elif model_name == "CNN":
            model = MNISTModelCNN()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "FashionMNIST":
        dataset = FashionMNISTDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "MLP":
            model = FashionMNISTModelMLP()
        elif model_name == "CNN":
            model = FashionMNISTModelCNN()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "EMNIST":
        dataset = EMNISTDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "MLP":
            model = EMNISTModelMLP()
        elif model_name == "CNN":
            model = EMNISTModelCNN()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "SYSCALL":
        dataset = SYSCALLDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "MLP":
            model = SyscallModelMLP()
        elif model_name == "SVM":
            model = SyscallModelSGDOneClassSVM()
        elif model_name == "Autoencoder":
            model = SyscallModelAutoencoder()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "CIFAR10":
        dataset = CIFAR10Dataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "ResNet9":
            model = CIFAR10ModelResNet(classifier="resnet9")
        elif model_name == "fastermobilenet":
            model = FasterMobileNet()
        elif model_name == "simplemobilenet":
            model = SimpleMobileNetV1()
        elif model_name == "CNN":
            model = CIFAR10ModelCNN()
        elif model_name == "CNNv2":
            model = CIFAR10ModelCNN_V2()
        elif model_name == "CNNv3":
            model = CIFAR10ModelCNN_V3()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "CIFAR100":
        dataset = CIFAR100Dataset(
            num_classes=100,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "CNN":
            model = CIFAR100ModelCNN()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "KITSUN":
        dataset = KITSUNDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        if model_name == "MLP":
            model = KitsunModelMLP()
        else:
            raise ValueError(f"Model {model} not supported for dataset {dataset_str}")
    elif dataset_str == "MilitarySAR":
        dataset = MilitarySARDataset(
            num_classes=10,
            partition_id=idx,
            partitions_number=n_nodes,
            iid=iid,
            partition=partition_selection,
            partition_parameter=partition_parameter,
            seed=42,
            config=config,
        )
        model = MilitarySARModelCNN()
    else:
        raise ValueError(f"Dataset {dataset_str} not supported")

    dataset = DataModule(
        train_set=dataset.train_set,
        train_set_indices=dataset.train_indices_map,
        test_set=dataset.test_set,
        test_set_indices=dataset.test_indices_map,
        local_test_set_indices=dataset.local_test_indices_map,
        num_workers=num_workers,
        partition_id=idx,
        partitions_number=n_nodes,
        batch_size=dataset.batch_size,
        label_flipping=label_flipping,
        data_poisoning=data_poisoning,
        poisoned_persent=poisoned_persent,
        poisoned_ratio=poisoned_ratio,
        targeted=targeted,
        target_label=target_label,
        target_changed_label=target_changed_label,
        noise_type=noise_type,
    )

    # - Import MNISTDatasetScikit (not torch component)
    # - Import scikit-learn model
    # - Import ScikitDataModule
    # - Import Scikit as trainer
    # - Import aggregation algorithm adapted to scikit-learn models (e.g. FedAvgSVM)

    trainer = None
    trainer_str = config.participant["training_args"]["trainer"]
    if trainer_str == "lightning":
        trainer = Lightning
    elif trainer_str == "scikit":
        raise NotImplementedError
    elif trainer_str == "siamese" and dataset_str == "CIFAR10":
        trainer = Siamese
        model = DualAggModel()
        config.participant["model_args"]["model"] = "DualAggModel"
        config.participant["data_args"]["dataset"] = "CIFAR10"
        config.participant["aggregator_args"]["algorithm"] = "DualHistAgg"
    else:
        raise ValueError(f"Trainer {trainer_str} not supported")

    if config.participant["device_args"]["malicious"]:
        node_cls = MaliciousNode
    else:
        if config.participant["device_args"]["role"] == Role.AGGREGATOR:
            node_cls = AggregatorNode
        elif config.participant["device_args"]["role"] == Role.TRAINER:
            node_cls = TrainerNode
        elif config.participant["device_args"]["role"] == Role.SERVER:
            node_cls = ServerNode
        elif config.participant["device_args"]["role"] == Role.IDLE:
            node_cls = IdleNode
        else:
            raise ValueError(f"Role {config.participant['device_args']['role']} not supported")

    VARIABILITY = 0.5

    def randomize_value(value, variability):
        min_value = max(0, value - variability)
        max_value = value + variability
        return random.uniform(min_value, max_value)

    config_keys = [
        ["reporter_args", "report_frequency"],
        ["discoverer_args", "discovery_frequency"],
        ["health_args", "health_interval"],
        ["health_args", "grace_time_health"],
        ["health_args", "check_alive_interval"],
        ["health_args", "send_alive_interval"],
        ["forwarder_args", "forwarder_interval"],
        ["forwarder_args", "forward_messages_interval"],
    ]

    for keys in config_keys:
        value = config.participant
        for key in keys[:-1]:
            value = value[key]
        value[keys[-1]] = randomize_value(value[keys[-1]], VARIABILITY)

    logging.info(f"Starting node {idx} with model {model_name}, trainer {trainer.__name__}, and as {node_cls.__name__}")

    node = node_cls(
        model=model,
        dataset=dataset,
        config=config,
        trainer=trainer,
        security=False,
        model_poisoning=model_poisoning,
        poisoned_ratio=poisoned_ratio,
        noise_type=noise_type,
    )
    await node.start_communications()
    await node.deploy_federation()

    # If it is an additional node, it should wait until additional_node_round to connect to the network
    # In order to do that, it should request the current round to the controller
    if additional_node_status:
        logging.info(f"Waiting for round {additional_node_round} to start")
        time.sleep(6000)  # DEBUG purposes
        import requests

        url = f"http://{node.config.participant['scenario_args']['controller']}/platform/{node.config.participant['scenario_args']['name']}/round"
        current_round = int(requests.get(url).json()["round"])
        while current_round < additional_node_round:
            logging.info(f"Waiting for round {additional_node_round} to start")
            time.sleep(10)
        logging.info(f"Round {additional_node_round} started, connecting to the network")

    if node.cm is not None:
        await node.cm.network_wait()


if __name__ == "__main__":
    config_path = str(sys.argv[1])
    config = Config(entity="participant", participant_config_file=config_path)
    if sys.platform == "win32" or config.participant["scenario_args"]["deployment"] == "docker":
        import asyncio

        asyncio.run(main(config), debug=False)
    else:
        try:
            import uvloop

            uvloop.run(main(config), debug=False)
        except ImportError:
            logging.warning("uvloop not available, using default loop")
            import asyncio

            asyncio.run(main(config), debug=False)



import json
import os
from collections.abc import Mapping
from functools import wraps

import requests
from eth_account import Account
from flask import Flask, jsonify, request
from retry import retry
from solcx import compile_standard, install_solc
from web3 import Web3
from web3.middleware import construct_sign_and_send_raw_middleware, geth_poa_middleware

app = Flask(__name__)


def error_handler(func):
    """Adds default status and header to all REST responses used for Oracle"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs), 200, {"Content-Type": "application/json"}
        except Exception as e:
            return jsonify({"error": str(e)}), 500, {"Content-Type": "application/json"}

    return wrapper


class Oracle:
    def __init__(self):
        # header file, required for interacting with chain code
        self.__contract_abi = dict()

        # stores gas expenses for experiments
        self.__gas_store = list()

        # stores timing records for experiments
        self.__time_store = list()

        # stores reputation records for experiments
        self.__reputation_store = list()

        # current (03.2024) average amount of WEI to pay for a unit of gas
        self.__gas_price_per_unit = 27.3

        # current (03.2024) average price in USD per WEI
        self.__price_USD_per_WEI = 0.00001971

        # static ip address of non-validator node (RPC)
        self.__blockchain_address = "http://172.25.0.104:8545"

        # executes RPC request to non-validator node until ready
        self.__ready = self.wait_for_blockchain()

        # creates an account from the primary key stored in the envs
        self.acc = self.__create_account()

        # create Web3 object for making transactions
        self.__web3 = self.__initialize_web3()

        # create a Web3 contract object from the compiled chaincode
        self.contract_obj = self.__compile_chaincode()

        # deploy the contract to the blockchain network
        self.__contract_address = self.deploy_chaincode()

        # update the contract object with the address
        self.contract_obj = self.__web3.eth.contract(
            abi=self.contract_obj.abi,
            bytecode=self.contract_obj.bytecode,
            address=self.contract_address,
        )

    @property
    def contract_abi(self):
        return self.__contract_abi

    @property
    def contract_address(self):
        return self.__contract_address

    @retry((Exception, requests.exceptions.HTTPError), tries=20, delay=10)
    def wait_for_blockchain(self) -> bool:
        """
        Executes REST post request for a selected RPC method to check if blockchain
        is up and running
        Returns: None

        """
        headers = {"Content-type": "application/json", "Accept": "application/json"}

        data = {"jsonrpc": "2.0", "method": "eth_accounts", "id": 1, "params": []}

        request = requests.post(url=self.__blockchain_address, json=data, headers=headers)

        # raise Exception if status is an error one
        request.raise_for_status()

        print("ORACLE: RPC node up and running", flush=True)

        return True

    def __initialize_web3(self):
        """
        Initializes Web3 object and configures it for PoA protocol
        Returns: Web3 object

        """

        # initialize Web3 object with ip of non-validator node
        web3 = Web3(Web3.HTTPProvider(self.__blockchain_address, request_kwargs={"timeout": 20}))  # 10

        # inject Proof-of-Authority settings to object
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        # automatically sign transactions if available for execution
        web3.middleware_onion.add(construct_sign_and_send_raw_middleware(self.acc))

        # inject local account as default
        web3.eth.default_account = self.acc.address

        # return initialized object for executing transaction
        print(f"SUCCESS: Account created at {self.acc.address}")
        return web3

    def __compile_chaincode(self):
        """
        Compile raw chaincode and create Web3 contract object with it
        Returns: Web3 contract object

        """

        # open raw solidity file
        with open("reputation_system.sol") as file:
            simple_storage_file = file.read()

        # set compiler version
        install_solc("0.8.22")

        # compile solidity code
        compiled_sol = compile_standard(
            {
                "language": "Solidity",
                "sources": {"reputation_system.sol": {"content": simple_storage_file}},
                "settings": {
                    "evmVersion": "paris",
                    "outputSelection": {"*": {"*": ["abi", "metadata", "evm.bytecode", "evm.sourceMap"]}},
                    "optimizer": {"enabled": True, "runs": 1000},
                },
            },
            solc_version="0.8.22",
        )

        # store compiled code as json
        with open("compiled_code.json", "w") as file:
            json.dump(compiled_sol, file)

        # retrieve bytecode from the compiled contract
        contract_bytecode = compiled_sol["contracts"]["reputation_system.sol"]["ReputationSystem"]["evm"]["bytecode"][
            "object"
        ]

        # retrieve ABI from compiled contract
        self.__contract_abi = json.loads(
            compiled_sol["contracts"]["reputation_system.sol"]["ReputationSystem"]["metadata"]
        )["output"]["abi"]

        print("Oracle: Solidity files compiled and bytecode ready", flush=True)

        # return draft Web3 contract object
        return self.__web3.eth.contract(abi=self.__contract_abi, bytecode=contract_bytecode)

    @staticmethod
    def __create_account():
        """
        Retrieves the private key from the envs, set during docker build
        Returns: Web3 account object

        """

        # retrieve private key, set during ducker build
        private_key = os.environ.get("PRIVATE_KEY")

        # return Web3 account object
        return Account.from_key("0x" + private_key)

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def transfer_funds(self, address):
        """
        Creates transaction to blockchain network for assigning funds to Cores
        Args:
            address: public wallet address of Core to assign funds to

        Returns: Transaction receipt

        """

        # create raw transaction with all required parameters to change state of ledger
        raw_transaction = {
            "chainId": self.__web3.eth.chain_id,
            "from": self.acc.address,
            "value": self.__web3.to_wei("500", "ether"),
            "to": self.__web3.to_checksum_address(address),
            "nonce": self.__web3.eth.get_transaction_count(self.acc.address, "pending"),
            "gasPrice": self.__web3.to_wei(self.__gas_price_per_unit, "gwei"),
            "gas": self.__web3.to_wei("22000", "wei"),
        }

        # sign transaction with private key and execute it
        tx_receipt = self.__sign_and_deploy(raw_transaction)

        # return transaction receipt
        return f"SUCESS: {tx_receipt}"

    def __sign_and_deploy(self, trx_hash):
        """
        Signs a function call to the chain code with the primary key and awaits the receipt
        Args:
            trx_hash: Transformed dictionary of all properties relevant for call to chain code

        Returns: transaction receipt confirming the successful write to the ledger

        """

        # transaction is signed with private key
        signed_transaction = self.__web3.eth.account.sign_transaction(trx_hash, private_key=self.acc.key)

        # confirmation that transaction was passed from non-validator node to validator nodes
        executed_transaction = self.__web3.eth.send_raw_transaction(signed_transaction.rawTransaction)

        # non-validator node awaited the successful validation by validation nodes and returns receipt
        transaction_receipt = self.__web3.eth.wait_for_transaction_receipt(executed_transaction, timeout=20)  # 5

        # report used gas for experiment
        self.report_gas(transaction_receipt.gasUsed, 0)

        return transaction_receipt

    @retry(Exception, tries=20, delay=5)
    def deploy_chaincode(self):
        """
        Creates transaction to deploy chain code on the blockchain network by
        sending transaction to non-validator node
        Returns: address of chain code on the network

        """

        # create raw transaction with all properties to deploy contract
        raw_transaction = self.contract_obj.constructor().build_transaction({
            "chainId": self.__web3.eth.chain_id,
            "from": self.acc.address,
            "value": self.__web3.to_wei("3", "ether"),
            "gasPrice": self.__web3.to_wei(self.__gas_price_per_unit, "gwei"),
            "nonce": self.__web3.eth.get_transaction_count(self.acc.address, "pending"),
        })

        # sign transaction with private key and executes it
        tx_receipt = self.__sign_and_deploy(raw_transaction)

        # store the address received from the non-validator node
        contract_address = tx_receipt["contractAddress"]

        # returns contract address to provide to the cores later
        return contract_address

    def get_balance(self, addr):
        """
        Creates transaction to blockchain network to request balance for parameter address
        Args:
            addr: public wallet address of account

        Returns: current balance in ether (ETH)

        """

        # converts address type required for making a transaction
        cAddr = self.__web3.to_checksum_address(addr)

        # executes the transaction directly, no signing required
        balance = self.__web3.eth.get_balance(cAddr, "pending")

        # returns JSON response with ether balance to requesting core
        return {"address": cAddr, "balance_eth": self.__web3.from_wei(balance, "ether")}

    def report_gas(self, amount: int, aggregation_round: int) -> None:
        """
        Experiment method for collecting and reporting gas usage statistics
        Args:
            aggregation_round: Aggregation round of sender
            amount: Amount of gas spent in WEI

        Returns: None

        """

        # store the recorded gas for experiment
        self.__gas_store.append((amount, aggregation_round))

    def get_gas_report(self) -> Mapping[str, str]:
        """
        Experiment method for requesting the summed up records of reported gas usage
        Returns: JSON with name:value (WEI/USD) for every reported node

        """
        # sum up all reported costs
        total_wei = sum(record[0] for record in self.__gas_store)

        # convert sum in WEI to USD by computing with gas price USD per WEI
        total_usd = round(total_wei * self.__price_USD_per_WEI)

        return {"Sum (WEI)": total_wei, "Sum (USD)": f"{total_usd:,}"}

    @property
    def gas_store(self):
        """
        Experiment method for requesting the detailed records of the gas reports
        Returns: list of records of type: list[(node, timestamp, gas)]

        """
        return self.__gas_store

    def report_time(self, time_s: float, aggregation_round: int) -> None:
        """
        Experiment method for collecting and reporting time statistics
        Args:
            aggregation_round: Aggregation round of node
            method: Name of node which reports time
            time_s: Amount of time spend on method

        Returns: None

        """

        # store the recorded time for experiment
        self.__time_store.append((time_s, aggregation_round))

    def report_reputation(self, records: list, aggregation_round: int, sender: str) -> None:
        """
        Experiment method for collecting and reporting reputations statistics
        Args:
            aggregation_round: Current aggregation round of sender
            records: list of (name:reputation) records
            sender: node reporting its local view

        Returns: None

        """

        # store the recorded reputation for experiment
        self.__reputation_store.extend([(record[0], record[1], aggregation_round, sender) for record in records])

    @property
    def time_store(self) -> list:
        """
        Experiment method for requesting all records of nodes which reported timings
        Returns: JSON with method:(sum_time, n_calls) for every reported node

        """
        return self.__time_store

    @property
    def reputation_store(self) -> list:
        """
        Experiment method for requesting all records of reputations
        Returns: list with (name, reputation, timestamp)

        """
        return self.__reputation_store

    @property
    def ready(self) -> bool:
        """
        Returns true if the Oracle is ready itself and the chain code was deployed successfully
        Returns: True if ready False otherwise

        """
        return self.__ready


@app.route("/")
@error_handler
def home():
    return jsonify({"Message": "Oracle up and running"})


@app.route("/faucet", methods=["POST"])
@error_handler
def rest_transfer_funds():
    address = request.get_json().get("address")
    return jsonify({"Message": oracle.transfer_funds(address)})


@app.route("/gas", methods=["POST"])
@error_handler
def rest_report_gas():
    amount = request.get_json().get("amount")
    aggregation_round = request.get_json().get("round")
    oracle.report_gas(amount, aggregation_round)
    return jsonify(oracle.get_gas_report())


@app.route("/balance", methods=["GET"])
@error_handler
def rest_get_balance():
    addr = request.get_json().get("address")
    return jsonify(oracle.get_balance(addr))


@app.route("/status", methods=["GET"])
@error_handler
def rest_status():
    if not oracle.ready:
        return {"message": "Blockchain does not respond, wait 10"}
    else:
        return {"message": "Blockchain responded"}


@app.route("/contract", methods=["GET"])
@error_handler
def rest_contract():
    return jsonify({"address": oracle.contract_address, "abi": oracle.contract_abi})


@app.route("/gas", methods=["GET"])
@error_handler
def rest_get_gas_report():
    return oracle.get_gas_report()


@app.route("/gas_series", methods=["GET"])
@error_handler
def rest_get_gas_series():
    return oracle.gas_store


@app.route("/time", methods=["POST"])
@error_handler
def rest_report_time():
    time = request.get_json().get("time")
    aggregation_round = request.get_json().get("round")
    oracle.report_time(time, aggregation_round)
    return jsonify({"Message": "Reported time successfully"})


@app.route("/time", methods=["GET"])
@error_handler
def rest_get_time_report():
    return oracle.time_store


@app.route("/reputation", methods=["POST"])
@error_handler
def rest_report_reputation():
    records = request.get_json().get("records")
    round = request.get_json().get("round")
    sender = request.get_json().get("sender")
    oracle.report_reputation(records, round, sender)
    return jsonify({"Message": "Reported reputation successfully"})


@app.route("/reputation", methods=["GET"])
@error_handler
def rest_get_reputation_timeseries():
    return oracle.reputation_store


if __name__ == "__main__":
    oracle = Oracle()
    app.run(debug=False, host="0.0.0.0", port=8081)



import logging
import math
import numbers
import os.path
import statistics
from datetime import datetime
from math import e
from os.path import exists

import numpy as np
import pandas as pd
import shap
import torch.nn
from art.estimators.classification import PyTorchClassifier
from art.metrics import clever_u
from codecarbon import EmissionsTracker
from scipy.stats import variation
from torch import nn, optim

dirname = os.path.dirname(__file__)
logger = logging.getLogger(__name__)

R_L1 = 40
R_L2 = 2
R_LI = 0.1


def get_mapped_score(score_key, score_map):
    """
    Finds the score by the score_key in the score_map.

    Args:
        score_key (string): The key to look up in the score_map.
        score_map (dict): The score map defined in the eval_metrics.json file.

    Returns:
        float: The normalized score of [0, 1].
    """
    score = 0
    if score_map is None:
        logger.warning("Score map is missing")
    else:
        keys = [key for key, value in score_map.items()]
        scores = [value for key, value in score_map.items()]
        normalized_scores = get_normalized_scores(scores)
        normalized_score_map = dict(zip(keys, normalized_scores, strict=False))
        score = normalized_score_map.get(score_key, np.nan)

    return score


def get_normalized_scores(scores):
    """
    Calculates the normalized scores of a list.

    Args:
        scores (list): The values that will be normalized.

    Returns:
        list: The normalized list.
    """
    normalized = [(x - np.min(scores)) / (np.max(scores) - np.min(scores)) for x in scores]
    return normalized


def get_range_score(value, ranges, direction="asc"):
    """
    Maps the value to a range and gets the score by the range and direction.

    Args:
        value (int): The input score.
        ranges (list): The ranges defined.
        direction (string): Asc means the higher the range the higher the score, desc means otherwise.

    Returns:
        float: The normalized score of [0, 1].
    """

    if not (type(value) == int or type(value) == float):
        logger.warning("Input value is not a number")
        logger.warning(f"{value}")
        return 0
    else:
        score = 0
        if ranges is None:
            logger.warning("Score ranges are missing")
        else:
            total_bins = len(ranges) + 1
            bin = np.digitize(value, ranges, right=True)
            score = 1 - (bin / total_bins) if direction == "desc" else bin / total_bins
        return score


def get_map_value_score(score_key, score_map):
    """
    Finds the score by the score_key in the score_map and returns the value.

    Args:
        score_key (string): The key to look up in the score_map.
        score_map (dict): The score map defined in the eval_metrics.json file.

    Returns:
        float: The score obtained in the score_map.
    """
    score = 0
    if score_map is None:
        logger.warning("Score map is missing")
    else:
        score = score_map[score_key]
    return score


def get_true_score(value, direction):
    """
    Returns the negative of the value if direction is 'desc', otherwise returns value.

    Args:
        value (int): The input score.
        direction (string): Asc means the higher the range the higher the score, desc means otherwise.

    Returns:
        float: The score obtained.
    """

    if value is True:
        return 1
    elif value is False:
        return 0
    else:
        if not (type(value) == int or type(value) == float):
            logger.warning("Input value is not a number")
            logger.warning(f"{value}.")
            return 0
        else:
            if direction == "desc":
                return 1 - value
            else:
                return value


def get_scaled_score(value, scale: list, direction: str):
    """
    Maps a score of a specific scale into the scale between zero and one.

    Args:
        value (int or float): The raw value of the metric.
        scale (list): List containing the minimum and maximum value the value can fall in between.

    Returns:
        float: The normalized score of [0, 1].
    """

    score = 0
    try:
        value_min, value_max = scale[0], scale[1]
    except Exception:
        logger.warning("Score minimum or score maximum is missing. The minimum has been set to 0 and the maximum to 1")
        value_min, value_max = 0, 1
    if not value:
        logger.warning("Score value is missing. Set value to zero")
    else:
        low, high = 0, 1
        if value >= value_max:
            score = 1
        elif value <= value_min:
            score = 0
        else:
            diff = value_max - value_min
            diffScale = high - low
            score = (float(value) - value_min) * (float(diffScale) / diff) + low
        if direction == "desc":
            score = high - score

    return score


def get_value(value):
    """
    Get the value of a metric.

    Args:
        value (float): The value of the metric.

    Returns:
        float: The value of the metric.
    """

    return value


def check_properties(*args):
    """
    Check if all the arguments have values.

    Args:
        args (list): All the arguments.

    Returns:
        float: The mean of arguments that have values.
    """

    result = map(lambda x: x is not None and x != "", args)
    return np.mean(list(result))


def get_cv(list=None, std=None, mean=None):
    """
    Get the coefficient of variation.

    Args:
        list (list): List in which the coefficient of variation will be calculated.
        std (float): Standard deviation of a list.
        mean (float): Mean of a list.

    Returns:
        float: The coefficient of variation calculated.
    """
    if std is not None and mean is not None:
        return std / mean

    if list is not None:
        return np.std(list) / np.mean(list)

    return 0


def get_global_privacy_risk(dp, epsilon, n):
    """
    Calculates the global privacy risk by epsilon and the number of clients.

    Args:
        dp (bool): Indicates if differential privacy is used or not.
        epsilon (int): The epsilon value.
        n (int): The number of clients in the scenario.

    Returns:
        float: The global privacy risk.
    """

    if dp is True and isinstance(epsilon, numbers.Number):
        return 1 / (1 + (n - 1) * math.pow(e, -epsilon))
    else:
        return 1


def get_elapsed_time(scenario):
    """
    Calculates the elapsed time during the execution of the scenario.

    Args:
        scenario (object): Scenario required.

    Returns:
        float: The elapsed time.
    """
    start_time = scenario[1]
    end_time = scenario[2]

    start_date = datetime.strptime(start_time, "%d/%m/%Y %H:%M:%S")
    end_date = datetime.strptime(end_time, "%d/%m/%Y %H:%M:%S")

    elapsed_time = (end_date - start_date).total_seconds() / 60

    return elapsed_time


def get_bytes_models(models_files):
    """
    Calculates the mean bytes of the final models of the nodes.

    Args:
        models_files (list): List of final models.

    Returns:
        float: The mean bytes of the models.
    """

    total_models_size = 0
    number_models = len(models_files)

    for file in models_files:
        model_size = os.path.getsize(file)
        total_models_size += model_size

    avg_model_size = total_models_size / number_models

    return avg_model_size


def get_bytes_sent_recv(bytes_sent_files, bytes_recv_files):
    """
    Calculates the mean bytes sent and received of the nodes.

    Args:
        bytes_sent_files (list): Files that contain the bytes sent of the nodes.
        bytes_recv_files (list): Files that contain the bytes received of the nodes.

    Returns:
        4-tupla: The total bytes sent, the total bytes received, the mean bytes sent and the mean bytes received of the nodes.
    """
    total_upload_bytes = 0
    total_download_bytes = 0
    number_files = len(bytes_sent_files)

    for file_bytes_sent, file_bytes_recv in zip(bytes_sent_files, bytes_recv_files, strict=False):
        with open(file_bytes_sent) as f:
            bytes_sent = f.read()

        with open(file_bytes_recv) as f:
            bytes_recv = f.read()

        total_upload_bytes += int(bytes_sent)
        total_download_bytes += int(bytes_recv)

    avg_upload_bytes = total_upload_bytes / number_files
    avg_download_bytes = total_download_bytes / number_files
    return (
        total_upload_bytes,
        total_download_bytes,
        avg_upload_bytes,
        avg_download_bytes,
    )


def get_avg_loss_accuracy(loss_files, accuracy_files):
    """
    Calculates the mean accuracy and loss models of the nodes.

    Args:
        loss_files (list): Files that contain the loss of the models of the nodes.
        accuracy_files (list): Files that contain the acurracies of the models of the nodes.

    Returns:
        3-tupla: The mean loss of the models, the mean accuracies of the models, the standard deviation of the accuracies of the models.
    """
    total_accuracy = 0
    total_loss = 0
    number_files = len(loss_files)
    accuracies = []

    for file_loss, file_accuracy in zip(loss_files, accuracy_files, strict=False):
        with open(file_loss) as f:
            loss = f.read()

        with open(file_accuracy) as f:
            accuracy = f.read()

        total_loss += float(loss)
        total_accuracy += float(accuracy)
        accuracies.append(float(accuracy))

    avg_loss = total_loss / number_files
    avg_accuracy = total_accuracy / number_files

    std_accuracy = statistics.stdev(accuracies)

    return avg_loss, avg_accuracy, std_accuracy


def get_feature_importance_cv(model, test_sample):
    """
    Calculates the coefficient of variation of the feature importance.

    Args:
        model (object): The model.
        test_sample (object): One test sample to calculate the feature importance.

    Returns:
        float: The coefficient of variation of the feature importance.
    """

    try:
        cv = 0
        batch_size = 10
        device = "cpu"

        if isinstance(model, torch.nn.Module):
            batched_data, _ = test_sample

            n = batch_size
            m = math.floor(0.8 * n)

            background = batched_data[:m].to(device)
            test_data = batched_data[m:n].to(device)

            e = shap.DeepExplainer(model, background)
            shap_values = e.shap_values(test_data)
            if shap_values is not None and len(shap_values) > 0:
                sums = np.array([shap_values[i].sum() for i in range(len(shap_values))])
                abs_sums = np.absolute(sums)
                cv = variation(abs_sums)
    except Exception as e:
        logger.warning("Could not compute feature importance CV with shap")
        cv = 1
    if math.isnan(cv):
        cv = 1
    return cv


def get_clever_score(model, test_sample, nb_classes, learning_rate):
    """
    Calculates the CLEVER score.

    Args:
        model (object): The model.
        test_sample (object): One test sample to calculate the CLEVER score.
        nb_classes (int): The nb_classes of the model.
        learning_rate (float): The learning rate of the model.

    Returns:
        float: The CLEVER score.
    """

    images, _ = test_sample
    background = images[-1]

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), learning_rate)

    # Create the ART classifier
    classifier = PyTorchClassifier(
        model=model,
        loss=criterion,
        optimizer=optimizer,
        input_shape=(1, 28, 28),
        nb_classes=nb_classes,
    )

    score_untargeted = clever_u(
        classifier,
        background.numpy(),
        10,
        5,
        R_L2,
        norm=2,
        pool_factor=3,
        verbose=False,
    )
    return score_untargeted


def stop_emissions_tracking_and_save(
    tracker: EmissionsTracker,
    outdir: str,
    emissions_file: str,
    role: str,
    workload: str,
    sample_size: int = 0,
):
    """
    Stops emissions tracking object from CodeCarbon and saves relevant information to emissions.csv file.

    Args:
        tracker (object): The emissions tracker object holding information.
        outdir (str): The path of the output directory of the experiment.
        emissions_file (str): The path to the emissions file.
        role (str): Either client or server depending on the role.
        workload (str): Either aggregation or training depending on the workload.
        sample_size (int): The number of samples used for training, if aggregation 0.
    """

    tracker.stop()

    emissions_file = os.path.join(outdir, emissions_file)

    if exists(emissions_file):
        df = pd.read_csv(emissions_file)
    else:
        df = pd.DataFrame(
            columns=[
                "role",
                "energy_grid",
                "emissions",
                "workload",
                "CPU_model",
                "GPU_model",
            ]
        )
    try:
        energy_grid = (tracker.final_emissions_data.emissions / tracker.final_emissions_data.energy_consumed) * 1000
        df = pd.concat(
            [
                df,
                pd.DataFrame({
                    "role": role,
                    "energy_grid": [energy_grid],
                    "emissions": [tracker.final_emissions_data.emissions],
                    "workload": workload,
                    "CPU_model": tracker.final_emissions_data.cpu_model
                    if tracker.final_emissions_data.cpu_model
                    else "None",
                    "GPU_model": tracker.final_emissions_data.gpu_model
                    if tracker.final_emissions_data.gpu_model
                    else "None",
                    "CPU_used": True if tracker.final_emissions_data.cpu_energy else False,
                    "GPU_used": True if tracker.final_emissions_data.gpu_energy else False,
                    "energy_consumed": tracker.final_emissions_data.energy_consumed,
                    "sample_size": sample_size,
                }),
            ],
            ignore_index=True,
        )
        df.to_csv(emissions_file, encoding="utf-8", index=False)
    except Exception as e:
        logger.warning(e)



import collections

import torch.nn as nn

_activations = {"relu": nn.ReLU, "relu6": nn.ReLU6, "leaky_relu": nn.LeakyReLU}


class BaseBlock(nn.Module):
    def __init__(self):
        super(BaseBlock, self).__init__()
        self._layer: nn.Sequential

    def forward(self, x):
        return self._layer(x)


class DenseBlock(BaseBlock):
    def __init__(self, shape, **params):
        super(DenseBlock, self).__init__()
        in_dims, out_dims = shape
        _seq = collections.OrderedDict([
            ("dense", nn.Linear(in_dims, out_dims)),
        ])
        _act_name = params.get("activation")
        if _act_name:
            _seq.update({_act_name: _activations[_act_name](inplace=True)})

        self._layer = nn.Sequential(_seq)

        w_init = params.get("w_init")
        idx = list(dict(self._layer.named_children()).keys()).index("dense")
        if w_init:
            w_init(self._layer[idx].weight)
        b_init = params.get("b_init")
        if b_init:
            b_init(self._layer[idx].bias)


class Conv2DBlock(BaseBlock):
    def __init__(self, shape, stride, padding="same", **params):
        super(Conv2DBlock, self).__init__()

        h, w, in_channels, out_channels = shape
        _seq = collections.OrderedDict([
            (
                "conv",
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=(h, w),
                    stride=stride,
                    padding=padding,
                ),
            )
        ])

        _bn = params.get("batch_norm")
        if _bn:
            _seq.update({"bn": nn.BatchNorm2d(out_channels)})

        _act_name = params.get("activation")
        if _act_name:
            _seq.update({_act_name: _activations[_act_name](inplace=True)})

        _max_pool = params.get("max_pool")
        if _max_pool:
            _kernel_size = params.get("max_pool_size", 2)
            _stride = params.get("max_pool_stride", _kernel_size)
            _seq.update({"max_pool": nn.MaxPool2d(kernel_size=_kernel_size, stride=_stride)})

        self._layer = nn.Sequential(_seq)

        w_init = params.get("w_init")
        idx = list(dict(self._layer.named_children()).keys()).index("conv")
        if w_init:
            w_init(self._layer[idx].weight)
        b_init = params.get("b_init")
        if b_init:
            b_init(self._layer[idx].bias)



import torch

from nebula.core.models.nebulamodel import NebulaModel


class CIFAR10ModelCNN_V3(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, 3, 32, 32)
        self.learning_rate = learning_rate
        self.criterion = torch.torch.nn.CrossEntropyLoss()
        self.layer1 = torch.nn.Sequential(
            torch.nn.Conv2d(input_channels, 32, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(32),
            torch.nn.ReLU(),
            torch.nn.Conv2d(32, 32, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(32),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2),
            torch.nn.Dropout(0.25),
        )

        self.layer2 = torch.nn.Sequential(
            torch.nn.Conv2d(32, 64, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(64),
            torch.nn.ReLU(),
            torch.nn.Conv2d(64, 64, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(64),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2),
            torch.nn.Dropout(0.25),
        )

        self.layer3 = torch.nn.Sequential(
            torch.nn.Conv2d(64, 128, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(128),
            torch.nn.ReLU(),
            torch.nn.Conv2d(128, 128, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(128),
            torch.nn.ReLU(),
            torch.nn.MaxPool2d(kernel_size=2, stride=2),
            torch.nn.Dropout(0.25),
        )

        self.fc_layer = torch.nn.Sequential(
            torch.nn.Linear(128 * 4 * 4, 512),
            torch.nn.ReLU(),
            torch.nn.Dropout(0.5),
            torch.nn.Linear(512, num_classes),
        )

    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = x.view(x.size(0), -1)  # Flatten the layer
        x = self.fc_layer(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate,
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )
        return optimizer



import logging
import os
import re
import signal
import subprocess
import sys
import textwrap
import threading
import time

import psutil
from dotenv import load_dotenv
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from nebula.addons.env import check_environment
from nebula.config.config import Config
from nebula.config.mender import Mender
from nebula.scenarios import ScenarioManagement
from nebula.tests import main as deploy_tests


# Setup controller logger
class TermEscapeCodeFormatter(logging.Formatter):
    def __init__(self, fmt=None, datefmt=None, style="%", validate=True):
        super().__init__(fmt, datefmt, style, validate)

    def format(self, record):
        escape_re = re.compile(r"\x1b\[[0-9;]*m")
        record.msg = re.sub(escape_re, "", str(record.msg))
        return super().format(record)


log_console_format = "[%(levelname)s] - %(asctime)s - Controller - %(message)s"
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.INFO)
# console_handler.setFormatter(logging.Formatter(log_console_format))
console_handler.setFormatter(TermEscapeCodeFormatter(log_console_format))
logging.basicConfig(
    level=logging.DEBUG,
    handlers=[
        console_handler,
    ],
)


# Detect ctrl+c and run killports
def signal_handler(sig, frame):
    Controller.stop()
    sys.exit(0)


signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGINT, signal_handler)


class NebulaEventHandler(PatternMatchingEventHandler):
    """
    NebulaEventHandler handles file system events for .sh scripts.

    This class monitors the creation, modification, and deletion of .sh scripts
    in a specified directory.
    """

    patterns = ["*.sh", "*.ps1"]

    def __init__(self):
        super(NebulaEventHandler, self).__init__()
        self.last_processed = {}
        self.timeout_ns = 5 * 1e9
        self.processing_files = set()
        self.lock = threading.Lock()

    def _should_process_event(self, src_path: str) -> bool:
        current_time_ns = time.time_ns()
        logging.info(f"Current time (ns): {current_time_ns}")
        with self.lock:
            if src_path in self.last_processed:
                logging.info(f"Last processed time for {src_path}: {self.last_processed[src_path]}")
                last_time = self.last_processed[src_path]
                if current_time_ns - last_time < self.timeout_ns:
                    return False
            self.last_processed[src_path] = current_time_ns
        return True

    def _is_being_processed(self, src_path: str) -> bool:
        with self.lock:
            if src_path in self.processing_files:
                logging.info(f"Skipping {src_path} as it is already being processed.")
                return True
            self.processing_files.add(src_path)
        return False

    def _processing_done(self, src_path: str):
        with self.lock:
            if src_path in self.processing_files:
                self.processing_files.remove(src_path)

    def on_created(self, event):
        """
        Handles the event when a file is created.
        """
        if event.is_directory:
            return
        src_path = event.src_path
        if not self._should_process_event(src_path):
            return
        if self._is_being_processed(src_path):
            return
        logging.info("File created: %s" % src_path)
        try:
            self.run_script(src_path)
        finally:
            self._processing_done(src_path)

    def on_deleted(self, event):
        """
        Handles the event when a file is deleted.
        """
        if event.is_directory:
            return
        src_path = event.src_path
        if not self._should_process_event(src_path):
            return
        if self._is_being_processed(src_path):
            return
        logging.info("File deleted: %s" % src_path)
        directory_script = os.path.dirname(src_path)
        pids_file = os.path.join(directory_script, "current_scenario_pids.txt")
        logging.info(f"Killing processes from {pids_file}")
        try:
            self.kill_script_processes(pids_file)
            os.remove(pids_file)
        except FileNotFoundError:
            logging.warning(f"{pids_file} not found.")
        except Exception as e:
            logging.exception(f"Error while killing processes: {e}")
        finally:
            self._processing_done(src_path)

    def run_script(self, script):
        try:
            logging.info(f"Running script: {script}")
            if script.endswith(".sh"):
                result = subprocess.run(["bash", script], capture_output=True, text=True)
                logging.info(f"Script output:\n{result.stdout}")
                if result.stderr:
                    logging.error(f"Script error:\n{result.stderr}")
            elif script.endswith(".ps1"):
                subprocess.Popen(
                    ["powershell", "-ExecutionPolicy", "Bypass", "-File", script],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=False,
                )
            else:
                logging.error("Unsupported script format.")
                return
        except Exception as e:
            logging.exception(f"Error while running script: {e}")

    def kill_script_processes(self, pids_file):
        try:
            with open(pids_file) as f:
                pids = f.readlines()
                for pid in pids:
                    try:
                        pid = int(pid.strip())
                        if psutil.pid_exists(pid):
                            process = psutil.Process(pid)
                            children = process.children(recursive=True)
                            logging.info(f"Forcibly killing process {pid} and {len(children)} child processes...")
                            for child in children:
                                try:
                                    logging.info(f"Forcibly killing child process {child.pid}")
                                    child.kill()
                                except psutil.NoSuchProcess:
                                    logging.warning(f"Child process {child.pid} already terminated.")
                                except Exception as e:
                                    logging.exception(f"Error while forcibly killing child process {child.pid}: {e}")
                            try:
                                logging.info(f"Forcibly killing main process {pid}")
                                process.kill()
                            except psutil.NoSuchProcess:
                                logging.warning(f"Process {pid} already terminated.")
                            except Exception as e:
                                logging.exception(f"Error while forcibly killing main process {pid}: {e}")
                        else:
                            logging.warning(f"PID {pid} does not exist.")
                    except ValueError:
                        logging.exception(f"Invalid PID value in file: {pid}")
                    except Exception as e:
                        logging.exception(f"Error while forcibly killing process {pid}: {e}")
        except FileNotFoundError:
            logging.exception(f"PID file not found: {pids_file}")
        except Exception as e:
            logging.exception(f"Error while reading PIDs from file: {e}")


class Controller:
    def __init__(self, args):
        self.scenario_name = args.scenario_name if hasattr(args, "scenario_name") else None
        self.start_date_scenario = None
        self.federation = args.federation if hasattr(args, "federation") else None
        self.topology = args.topology if hasattr(args, "topology") else None
        self.waf_port = args.wafport if hasattr(args, "wafport") else 6000
        self.frontend_port = args.webport if hasattr(args, "webport") else 6060
        self.grafana_port = args.grafanaport if hasattr(args, "grafanaport") else 6040
        self.loki_port = args.lokiport if hasattr(args, "lokiport") else 6010
        self.statistics_port = args.statsport if hasattr(args, "statsport") else 8080
        self.simulation = args.simulation
        self.config_dir = args.config
        self.test = args.test if hasattr(args, "test") else False
        self.log_dir = args.logs
        self.cert_dir = args.certs
        self.env_path = args.env
        self.production = args.production if hasattr(args, "production") else False
        self.advanced_analytics = args.advanced_analytics if hasattr(args, "advanced_analytics") else False
        self.matrix = args.matrix if hasattr(args, "matrix") else None
        self.root_path = (
            args.root_path
            if hasattr(args, "root_path")
            else os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        )
        self.host_platform = "windows" if sys.platform == "win32" else "unix"

        # Network configuration (nodes deployment in a network)
        self.network_subnet = args.network_subnet if hasattr(args, "network_subnet") else None
        self.network_gateway = args.network_gateway if hasattr(args, "network_gateway") else None

        self.config = Config(entity="controller")
        self.topologymanager = None
        self.n_nodes = 0
        self.mender = None if self.simulation else Mender()
        self.use_blockchain = args.use_blockchain if hasattr(args, "use_blockchain") else False
        self.gpu_available = False

    def start(self):
        banner = """
                        ███╗   ██╗███████╗██████╗ ██╗   ██╗██╗      █████╗
                        ████╗  ██║██╔════╝██╔══██╗██║   ██║██║     ██╔══██╗
                        ██╔██╗ ██║█████╗  ██████╔╝██║   ██║██║     ███████║
                        ██║╚██╗██║██╔══╝  ██╔══██╗██║   ██║██║     ██╔══██║
                        ██║ ╚████║███████╗██████╔╝╚██████╔╝███████╗██║  ██║
                        ╚═╝  ╚═══╝╚══════╝╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝
                          A Platform for Decentralized Federated Learning
                            Created by Enrique Tomás Martínez Beltrán
                              https://github.com/CyberDataLab/nebula
                    """
        print("\x1b[0;36m" + banner + "\x1b[0m")

        # Load the environment variables
        load_dotenv(self.env_path)

        # Save controller pid
        with open(os.path.join(os.path.dirname(__file__), "controller.pid"), "w") as f:
            f.write(str(os.getpid()))

        # Check information about the environment
        check_environment()

        # Save the configuration in environment variables
        logging.info("Saving configuration in environment variables...")
        os.environ["NEBULA_ROOT"] = self.root_path
        os.environ["NEBULA_LOGS_DIR"] = self.log_dir
        os.environ["NEBULA_CONFIG_DIR"] = self.config_dir
        os.environ["NEBULA_CERTS_DIR"] = self.cert_dir
        os.environ["NEBULA_STATISTICS_PORT"] = str(self.statistics_port)
        os.environ["NEBULA_ROOT_HOST"] = self.root_path
        os.environ["NEBULA_HOST_PLATFORM"] = self.host_platform

        if self.production:
            self.run_waf()
            logging.info(f"NEBULA WAF is running at port {self.waf_port}")
            logging.info(f"Grafana Dashboard is running at port {self.grafana_port}")

        if self.test:
            self.run_test()
        else:
            self.run_frontend()
            logging.info(f"NEBULA Frontend is running at port {self.frontend_port}")

        # Watchdog for running additional scripts in the host machine (i.e. during the execution of a federation)
        event_handler = NebulaEventHandler()
        observer = Observer()
        observer.schedule(event_handler, path=self.config_dir, recursive=False)
        observer.start()

        if self.mender:
            logging.info("[Mender.module] Mender module initialized")
            time.sleep(2)
            mender = Mender()
            logging.info("[Mender.module] Getting token from Mender server: {}".format(os.getenv("MENDER_SERVER")))
            mender.renew_token()
            time.sleep(2)
            logging.info(
                "[Mender.module] Getting devices from {} with group Cluster_Thun".format(os.getenv("MENDER_SERVER"))
            )
            time.sleep(2)
            devices = mender.get_devices_by_group("Cluster_Thun")
            logging.info("[Mender.module] Getting a pool of devices: 5 devices")
            # devices = devices[:5]
            for i in self.config.participants:
                logging.info(
                    "[Mender.module] Device {} | IP: {}".format(i["device_args"]["idx"], i["network_args"]["ip"])
                )
                logging.info("[Mender.module] \tCreating artifacts...")
                logging.info("[Mender.module] \tSending NEBULA Core...")
                # mender.deploy_artifact_device("my-update-2.0.mender", i['device_args']['idx'])
                logging.info("[Mender.module] \tSending configuration...")
                time.sleep(5)
            sys.exit(0)

        logging.info("Press Ctrl+C for exit from NEBULA (global exit)")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            logging.info("Closing NEBULA (exiting from components)... Please wait")
            observer.stop()
            self.stop()

        observer.join()

    def run_waf(self):
        docker_compose_template = textwrap.dedent(
            """
            services:
            {}
        """
        )

        waf_template = textwrap.dedent(
            """
            nebula-waf:
                container_name: nebula-waf
                image: nebula-waf
                build:
                    context: .
                    dockerfile: Dockerfile-waf
                restart: unless-stopped
                volumes:
                    - {log_path}/waf/nginx:/var/log/nginx
                extra_hosts:
                    - "host.docker.internal:host-gateway"
                ipc: host
                privileged: true
                ports:
                    - {waf_port}:80
                networks:
                    nebula-net-base:
                        ipv4_address: {ip}
        """
        )

        grafana_template = textwrap.dedent(
            """
            grafana:
                container_name: nebula-waf-grafana
                image: nebula-waf-grafana
                build:
                    context: .
                    dockerfile: Dockerfile-grafana
                restart: unless-stopped
                environment:
                    - GF_SECURITY_ADMIN_PASSWORD=admin
                    - GF_USERS_ALLOW_SIGN_UP=false
                    - GF_SERVER_HTTP_PORT=3000
                    - GF_SERVER_PROTOCOL=http
                    - GF_SERVER_DOMAIN=localhost:{grafana_port}
                    - GF_SERVER_ROOT_URL=http://localhost:{grafana_port}/grafana/
                    - GF_SERVER_SERVE_FROM_SUB_PATH=true
                    - GF_DASHBOARDS_DEFAULT_HOME_DASHBOARD_PATH=/var/lib/grafana/dashboards/dashboard.json
                    - GF_METRICS_MAX_LIMIT_TSDB=0
                ports:
                    - {grafana_port}:3000
                ipc: host
                privileged: true
                networks:
                    nebula-net-base:
                        ipv4_address: {ip}
        """
        )

        loki_template = textwrap.dedent(
            """
            loki:
                container_name: nebula-waf-loki
                image: nebula-waf-loki
                build:
                    context: .
                    dockerfile: Dockerfile-loki
                restart: unless-stopped
                volumes:
                    - ./loki-config.yml:/mnt/config/loki-config.yml
                ports:
                    - {loki_port}:3100
                user: "0:0"
                command:
                    - '-config.file=/mnt/config/loki-config.yml'
                networks:
                    nebula-net-base:
                        ipv4_address: {ip}
        """
        )

        promtail_template = textwrap.dedent(
            """
            promtail:
                container_name: nebula-waf-promtail
                image: nebula-waf-promtail
                build:
                    context: .
                    dockerfile: Dockerfile-promtail
                restart: unless-stopped
                volumes:
                    - {log_path}/waf/nginx:/var/log/nginx
                    - ./promtail-config.yml:/etc/promtail/config.yml
                command:
                    - '-config.file=/etc/promtail/config.yml'
                networks:
                    nebula-net-base:
                        ipv4_address: {ip}
        """
        )

        waf_template = textwrap.indent(waf_template, " " * 4)
        grafana_template = textwrap.indent(grafana_template, " " * 4)
        loki_template = textwrap.indent(loki_template, " " * 4)
        promtail_template = textwrap.indent(promtail_template, " " * 4)

        network_template = textwrap.dedent(
            """
            networks:
                nebula-net-base:
                    name: nebula-net-base
                    driver: bridge
                    ipam:
                        config:
                            - subnet: {}
                              gateway: {}
        """
        )

        # Generate the Docker Compose file dynamically
        services = ""
        services += waf_template.format(
            path=self.root_path,
            log_path=os.environ["NEBULA_LOGS_DIR"],
            waf_port=self.waf_port,
            gw="192.168.10.1",
            ip="192.168.10.200",
        )

        services += grafana_template.format(
            log_path=os.environ["NEBULA_LOGS_DIR"],
            grafana_port=self.grafana_port,
            loki_port=self.loki_port,
            ip="192.168.10.201",
        )

        services += loki_template.format(loki_port=self.loki_port, ip="192.168.10.202")

        services += promtail_template.format(log_path=os.environ["NEBULA_LOGS_DIR"], ip="192.168.10.203")

        docker_compose_file = docker_compose_template.format(services)
        docker_compose_file += network_template.format("192.168.10.0/24", "192.168.10.1")

        # Write the Docker Compose file in waf directory
        with open(
            f"{os.path.join(os.environ['NEBULA_ROOT'], 'nebula', 'addons', 'waf', 'docker-compose.yml')}",
            "w",
        ) as f:
            f.write(docker_compose_file)

        # Start the Docker Compose file, catch error if any
        try:
            subprocess.check_call([
                "docker",
                "compose",
                "-f",
                f"{os.path.join(os.environ['NEBULA_ROOT'], 'nebula', 'addons', 'waf', 'docker-compose.yml')}",
                "up",
                "--build",
                "-d",
            ])
        except subprocess.CalledProcessError:
            raise Exception(
                "Docker Compose failed to start, please check if Docker Compose is installed (https://docs.docker.com/compose/install/) and Docker Engine is running."
            )

    def run_frontend(self):
        if sys.platform == "win32":
            if not os.path.exists("//./pipe/docker_Engine"):
                raise Exception(
                    "Docker is not running, please check if Docker is running and Docker Compose is installed."
                )
        else:
            if not os.path.exists("/var/run/docker.sock"):
                raise Exception(
                    "/var/run/docker.sock not found, please check if Docker is running and Docker Compose is installed."
                )

        docker_compose_template = textwrap.dedent(
            """
            services:
            {}
        """
        )

        frontend_template = textwrap.dedent(
            """
            nebula-frontend:
                container_name: nebula-frontend
                image: nebula-frontend
                build:
                    context: {path}
                    dockerfile: nebula/frontend/Dockerfile
                restart: unless-stopped
                volumes:
                    - {path}:/nebula
                    - /var/run/docker.sock:/var/run/docker.sock
                    - ./config/nebula:/etc/nginx/sites-available/default
                environment:
                    - NEBULA_PRODUCTION={production}
                    - NEBULA_GPU_AVAILABLE={gpu_available}
                    - NEBULA_ADVANCED_ANALYTICS={advanced_analytics}
                    - SERVER_LOG=/nebula/app/logs/server.log
                    - NEBULA_LOGS_DIR=/nebula/app/logs/
                    - NEBULA_CONFIG_DIR=/nebula/app/config/
                    - NEBULA_CERTS_DIR=/nebula/app/certs/
                    - NEBULA_ENV_PATH=/nebula/app/.env
                    - NEBULA_ROOT_HOST={path}
                    - NEBULA_HOST_PLATFORM={platform}
                    - NEBULA_DEFAULT_USER=admin
                    - NEBULA_DEFAULT_PASSWORD=admin
                    - NEBULA_FRONTEND_PORT={frontend_port}
                extra_hosts:
                    - "host.docker.internal:host-gateway"
                ipc: host
                privileged: true
                ports:
                    - {frontend_port}:80
                    - {statistics_port}:8080
                networks:
                    nebula-net-base:
                        ipv4_address: {ip}
        """
        )
        frontend_template = textwrap.indent(frontend_template, " " * 4)

        network_template = textwrap.dedent(
            """
            networks:
                nebula-net-base:
                    name: nebula-net-base
                    driver: bridge
                    ipam:
                        config:
                            - subnet: {}
                              gateway: {}
        """
        )

        network_template_external = textwrap.dedent(
            """
            networks:
                nebula-net-base:
                    external: true
        """
        )

        try:
            subprocess.check_call(["nvidia-smi"])
            self.gpu_available = True
        except Exception:
            logging.info("No GPU available for the frontend, nodes will be deploy in CPU mode")

        # Generate the Docker Compose file dynamically
        services = ""
        services += frontend_template.format(
            production=self.production,
            gpu_available=self.gpu_available,
            advanced_analytics=self.advanced_analytics,
            path=self.root_path,
            platform=self.host_platform,
            gw="192.168.10.1",
            ip="192.168.10.100",
            frontend_port=self.frontend_port,
            statistics_port=self.statistics_port,
        )
        docker_compose_file = docker_compose_template.format(services)

        if self.production:
            # If WAF is enabled, we need to use the same network
            docker_compose_file += network_template_external
        else:
            docker_compose_file += network_template.format("192.168.10.0/24", "192.168.10.1")
        # Write the Docker Compose file in config directory
        with open(
            f"{os.path.join(os.environ['NEBULA_ROOT'], 'nebula', 'frontend', 'docker-compose.yml')}",
            "w",
        ) as f:
            f.write(docker_compose_file)

        # Start the Docker Compose file, catch error if any
        try:
            subprocess.check_call([
                "docker",
                "compose",
                "-f",
                f"{os.path.join(os.environ['NEBULA_ROOT'], 'nebula', 'frontend', 'docker-compose.yml')}",
                "up",
                "--build",
                "-d",
            ])
        except subprocess.CalledProcessError:
            raise Exception(
                "Docker Compose failed to start, please check if Docker Compose is installed (https://docs.docker.com/compose/install/) and Docker Engine is running."
            )

        except Exception as e:
            raise Exception(f"Error while starting the frontend: {e}")

    def run_test(self):
        deploy_tests.start()

    @staticmethod
    def stop_frontend():
        if sys.platform == "win32":
            try:
                # kill all the docker containers which contain the word "nebula"
                commands = [
                    """docker kill $(docker ps -q --filter ancestor=nebula-frontend) | Out-Null""",
                    """docker rm $(docker ps -a -q --filter ancestor=nebula-frontend) | Out-Null""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(f'powershell.exe -Command "{command}"')
                    # logging.info(f"Windows Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")
        else:
            try:
                commands = [
                    """docker kill $(docker ps -q --filter ancestor=nebula-frontend) > /dev/null 2>&1""",
                    """docker rm $(docker ps -a -q --filter ancestor=nebula-frontend) > /dev/null 2>&1""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(command)
                    # logging.info(f"Linux Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")

    @staticmethod
    def stop_network():
        if sys.platform == "win32":
            try:
                # kill all the docker containers which contain the word "nebula"
                commands = [
                    r"""docker network rm $(docker network ls | Where-Object { ($_ -split '\s+')[1] -like 'nebula-net-base' } | ForEach-Object { ($_ -split '\s+')[0] }) | Out-Null"""
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(f'powershell.exe -Command "{command}"')
                    # logging.info(f"Windows Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")
        else:
            try:
                commands = [
                    """docker network rm $(docker network ls | grep nebula-net-base | awk '{print $1}') > /dev/null 2>&1"""
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(command)
                    # logging.info(f"Linux Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")

    @staticmethod
    def stop_waf():
        if sys.platform == "win32":
            try:
                # kill all the docker containers which contain the word "nebula"
                commands = [
                    """docker compose -p waf down | Out-Null""",
                    """docker compose -p waf rm | Out-Null""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(f'powershell.exe -Command "{command}"')
                    # logging.info(f"Windows Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")
        else:
            try:
                commands = [
                    """docker compose -p waf down > /dev/null 2>&1""",
                    """docker compose -p waf rm > /dev/null 2>&1""",
                ]

                for command in commands:
                    time.sleep(1)
                    exit_code = os.system(command)
                    # logging.info(f"Linux Command '{command}' executed with exit code: {exit_code}")

            except Exception as e:
                raise Exception(f"Error while killing docker containers: {e}")

    @staticmethod
    def stop():
        logging.info("Closing NEBULA (exiting from components)... Please wait")
        ScenarioManagement.stop_participants()
        ScenarioManagement.stop_blockchain()
        Controller.stop_frontend()
        Controller.stop_waf()
        Controller.stop_network()
        controller_pid_file = os.path.join(os.path.dirname(__file__), "controller.pid")
        try:
            with open(controller_pid_file) as f:
                pid = int(f.read())
                os.kill(pid, signal.SIGKILL)
                os.remove(controller_pid_file)
        except Exception as e:
            logging.exception(f"Error while killing controller process: {e}")
        sys.exit(0)



import json
import logging
import os
import sys
from datetime import datetime

import docker

# Constants
TIMEOUT = 3600


# Detect CTRL+C from parent process
def signal_handler(signal, frame):
    logging.info("You pressed Ctrl+C [test]!")
    sys.exit(0)


# Create nebula netbase if it does not exist
def create_docker_network():
    client = docker.from_env()

    try:
        ipam_pool = docker.types.IPAMPool(subnet="192.168.10.0/24", gateway="192.168.10.1")
        ipam_config = docker.types.IPAMConfig(pool_configs=[ipam_pool])

        client.networks.create("nebula-net-base", driver="bridge", ipam=ipam_config)
        print("Docker network created successfully.")
    except docker.errors.APIError as e:
        print(f"Error creating Docker network: {e}")


# To add a new test create the option in the menu and a [test].json file in tests folder
def menu():
    # clear terminal
    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")

    banner = """
███╗   ██╗███████╗██████╗ ██╗   ██╗██╗      █████╗
████╗  ██║██╔════╝██╔══██╗██║   ██║██║     ██╔══██╗
██╔██╗ ██║█████╗  ██████╔╝██║   ██║██║     ███████║
██║╚██╗██║██╔══╝  ██╔══██╗██║   ██║██║     ██╔══██║
██║ ╚████║███████╗██████╔╝╚██████╔╝███████╗██║  ██║
╚═╝  ╚═══╝╚══════╝╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝
  A Platform for Decentralized Federated Learning
    Created by Enrique Tomás Martínez Beltrán
      https://github.com/CyberDataLab/nebula
                """
    print("\x1b[0;36m" + banner + "\x1b[0m")

    options = """
[1] Aggregation test
[2] Topology test
[3] Dataset test
[4] Attacks test
[5] Custom test
CTRL + C to exit
                """

    while True:
        print("\x1b[0;36m" + options + "\x1b[0m")
        selectedOption = input("\x1b[0;36m" + "> " + "\x1b[0m")
        if selectedOption == "1":
            run_test(os.path.join(os.path.dirname(__file__), "aggregation.json"))
        elif selectedOption == "2":
            run_test(os.path.join(os.path.dirname(__file__), "topology.json"))
        elif selectedOption == "3":
            run_test(os.path.join(os.path.dirname(__file__), "datasets.json"))
        elif selectedOption == "4":
            run_test(os.path.join(os.path.dirname(__file__), "attacks.json"))
        elif selectedOption == "5":
            run_test(os.path.join(os.path.dirname(__file__), "custom.json"))
        else:
            print("Choose a valid option")


# Check for error logs
def check_error_logs(test_name, scenario_name):
    try:
        log_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "app", "logs"))
        current_log = os.path.join(log_dir, scenario_name)
        test_dir = os.path.normpath(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "app", "tests")
        )
        output_log_path = os.path.join(test_dir, test_name + ".log")

        if not os.path.exists(test_dir):
            try:
                os.mkdir(test_dir)
            except Exception as e:
                logging.exception(f"Error creating test directory: {e}")

        with open(output_log_path, "a", encoding="utf-8") as f:
            f.write(f"Scenario: {scenario_name}\n")

            for log_file in os.listdir(current_log):
                if log_file.endswith("_error.log"):
                    log_file_path = os.path.join(current_log, log_file)
                    try:
                        with open(log_file_path, encoding="utf-8") as file:
                            content = file.read().strip()
                            if content:
                                f.write(f"{log_file} ❌ Errors found:\n{content}\n")
                            else:
                                f.write(f"{log_file} ✅ No errors found\n")
                    except Exception as e:
                        f.write(f"Error reading {log_file}: {e}\n")

            f.write("-" * os.get_terminal_size().columns + "\n")

    except Exception as e:
        print(f"Failed to write to log file {test_name + '.log'}: {e}")

    return output_log_path


# Load test from .json file
def load_test(test_path):
    with open(test_path, encoding="utf-8") as file:
        scenarios = json.load(file)
    return scenarios


# Run selected test
def run_test(test_path):
    test_name = f"test_nebula_{os.path.splitext(os.path.basename(test_path))[0]}_" + datetime.now().strftime(
        "%d_%m_%Y_%H_%M_%S"
    )

    for scenario in load_test(test_path):
        scenarioManagement = run_scenario(scenario)
        finished = scenarioManagement.scenario_finished(TIMEOUT)

        if finished:
            test_log_path = check_error_logs(test_name, scenarioManagement.scenario_name)
        else:
            test_dir = os.path.normpath(
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    "..",
                    "..",
                    "app",
                    "tests",
                )
            )
            output_log_path = os.path.join(test_dir, test_name + ".log")

            if not os.path.exists(test_dir):
                try:
                    os.mkdir(test_dir)
                except Exception as e:
                    logging.exception(f"Error creating test directory: {e}")

            try:
                with open(output_log_path, "a", encoding="utf-8") as f:
                    f.write(f"Scenario: {scenarioManagement.scenario_name} \n")
                    f.write("🕒❌ Timeout reached \n")
                    f.write("-" * os.get_terminal_size().columns + "\n")
            except Exception as e:
                print(f"Failed to write to log file {test_name + '.log'}: {e}")
            pass

    print("Results:")
    try:
        with open(test_log_path, encoding="utf-8") as f:
            print(f.read())
    except Exception as e:
        print(f"Failed to read the log file {test_name + '.log'}: {e}")


# Run a single scenario
def run_scenario(scenario):
    import subprocess

    from nebula.scenarios import ScenarioManagement

    # Manager for the actual scenario
    scenarioManagement = ScenarioManagement(scenario, "nebula-test")

    # Run the actual scenario
    try:
        if scenarioManagement.scenario.mobility:
            additional_participants = scenario["additional_participants"]
            schema_additional_participants = scenario["schema_additional_participants"]
            scenarioManagement.load_configurations_and_start_nodes(
                additional_participants, schema_additional_participants
            )
        else:
            scenarioManagement.load_configurations_and_start_nodes()
    except subprocess.CalledProcessError as e:
        logging.exception(f"Error docker-compose up: {e}")

    return scenarioManagement


if __name__ == "__main__":
    create_docker_network()
    menu()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nebula.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0cnebula.proto\x12\x06nebula"\xe4\x02\n\x07Wrapper\x12\x0e\n\x06source\x18\x01 \x01(\t\x12\x35\n\x11\x64iscovery_message\x18\x02 \x01(\x0b\x32\x18.nebula.DiscoveryMessageH\x00\x12\x31\n\x0f\x63ontrol_message\x18\x03 \x01(\x0b\x32\x16.nebula.ControlMessageH\x00\x12\x37\n\x12\x66\x65\x64\x65ration_message\x18\x04 \x01(\x0b\x32\x19.nebula.FederationMessageH\x00\x12-\n\rmodel_message\x18\x05 \x01(\x0b\x32\x14.nebula.ModelMessageH\x00\x12\x37\n\x12\x63onnection_message\x18\x06 \x01(\x0b\x32\x19.nebula.ConnectionMessageH\x00\x12\x33\n\x10response_message\x18\x07 \x01(\x0b\x32\x17.nebula.ResponseMessageH\x00\x42\t\n\x07message"\x9e\x01\n\x10\x44iscoveryMessage\x12/\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32\x1f.nebula.DiscoveryMessage.Action\x12\x10\n\x08latitude\x18\x02 \x01(\x02\x12\x11\n\tlongitude\x18\x03 \x01(\x02"4\n\x06\x41\x63tion\x12\x0c\n\x08\x44ISCOVER\x10\x00\x12\x0c\n\x08REGISTER\x10\x01\x12\x0e\n\nDEREGISTER\x10\x02"\x9a\x01\n\x0e\x43ontrolMessage\x12-\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32\x1d.nebula.ControlMessage.Action\x12\x0b\n\x03log\x18\x02 \x01(\t"L\n\x06\x41\x63tion\x12\t\n\x05\x41LIVE\x10\x00\x12\x0c\n\x08OVERHEAD\x10\x01\x12\x0c\n\x08MOBILITY\x10\x02\x12\x0c\n\x08RECOVERY\x10\x03\x12\r\n\tWEAK_LINK\x10\x04"\xcd\x01\n\x11\x46\x65\x64\x65rationMessage\x12\x30\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32 .nebula.FederationMessage.Action\x12\x11\n\targuments\x18\x02 \x03(\t\x12\r\n\x05round\x18\x03 \x01(\x05"d\n\x06\x41\x63tion\x12\x14\n\x10\x46\x45\x44\x45RATION_START\x10\x00\x12\x0e\n\nREPUTATION\x10\x01\x12\x1e\n\x1a\x46\x45\x44\x45RATION_MODELS_INCLUDED\x10\x02\x12\x14\n\x10\x46\x45\x44\x45RATION_READY\x10\x03"A\n\x0cModelMessage\x12\x12\n\nparameters\x18\x01 \x01(\x0c\x12\x0e\n\x06weight\x18\x02 \x01(\x03\x12\r\n\x05round\x18\x03 \x01(\x05"l\n\x11\x43onnectionMessage\x12\x30\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32 .nebula.ConnectionMessage.Action"%\n\x06\x41\x63tion\x12\x0b\n\x07\x43ONNECT\x10\x00\x12\x0e\n\nDISCONNECT\x10\x01"#\n\x0fResponseMessage\x12\x10\n\x08response\x18\x01 \x01(\tb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "nebula_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_WRAPPER"]._serialized_start = 25
    _globals["_WRAPPER"]._serialized_end = 381
    _globals["_DISCOVERYMESSAGE"]._serialized_start = 384
    _globals["_DISCOVERYMESSAGE"]._serialized_end = 542
    _globals["_DISCOVERYMESSAGE_ACTION"]._serialized_start = 490
    _globals["_DISCOVERYMESSAGE_ACTION"]._serialized_end = 542
    _globals["_CONTROLMESSAGE"]._serialized_start = 545
    _globals["_CONTROLMESSAGE"]._serialized_end = 699
    _globals["_CONTROLMESSAGE_ACTION"]._serialized_start = 623
    _globals["_CONTROLMESSAGE_ACTION"]._serialized_end = 699
    _globals["_FEDERATIONMESSAGE"]._serialized_start = 702
    _globals["_FEDERATIONMESSAGE"]._serialized_end = 907
    _globals["_FEDERATIONMESSAGE_ACTION"]._serialized_start = 807
    _globals["_FEDERATIONMESSAGE_ACTION"]._serialized_end = 907
    _globals["_MODELMESSAGE"]._serialized_start = 909
    _globals["_MODELMESSAGE"]._serialized_end = 974
    _globals["_CONNECTIONMESSAGE"]._serialized_start = 976
    _globals["_CONNECTIONMESSAGE"]._serialized_end = 1084
    _globals["_CONNECTIONMESSAGE_ACTION"]._serialized_start = 1047
    _globals["_CONNECTIONMESSAGE_ACTION"]._serialized_end = 1084
    _globals["_RESPONSEMESSAGE"]._serialized_start = 1086
    _globals["_RESPONSEMESSAGE"]._serialized_end = 1121
# @@protoc_insertion_point(module_scope)



import hashlib
import io
import logging
import traceback
from collections import OrderedDict

import torch
from lightning import Trainer
from lightning.pytorch.callbacks import RichModelSummary, RichProgressBar
from lightning.pytorch.callbacks.progress.rich_progress import RichProgressBarTheme

from nebula.core.utils.deterministic import enable_deterministic


class Siamese:
    def __init__(self, model, data, config=None, logger=None):
        # self.model = torch.compile(model, mode="reduce-overhead")
        self.model = model
        self.data = data
        self.config = config
        self.logger = logger
        self.__trainer = None
        self.epochs = 1
        logging.getLogger("lightning.pytorch").setLevel(logging.INFO)
        self.round = 0
        enable_deterministic(self.config)
        self.logger.log_data({"Round": self.round}, step=self.logger.global_step)

    @property
    def logger(self):
        return self._logger

    def get_round(self):
        return self.round

    def set_model(self, model):
        self.model = model

    def set_data(self, data):
        self.data = data

    def create_trainer(self):
        logging.info(
            "[Trainer] Creating trainer with accelerator: {}".format(
                self.config.participant["device_args"]["accelerator"]
            )
        )
        progress_bar = RichProgressBar(
            theme=RichProgressBarTheme(
                description="green_yellow",
                progress_bar="green1",
                progress_bar_finished="green1",
                progress_bar_pulse="#6206E0",
                batch_progress="green_yellow",
                time="grey82",
                processing_speed="grey82",
                metrics="grey82",
            ),
            leave=True,
        )
        if self.config.participant["device_args"]["accelerator"] == "gpu":
            # NEBULA uses 2 GPUs (max) to distribute the nodes.
            if self.config.participant["device_args"]["devices"] > 1:
                # If you have more than 2 GPUs, you should specify which ones to use.
                gpu_id = ([1] if self.config.participant["device_args"]["idx"] % 2 == 0 else [2],)
            else:
                # If there is only one GPU, it will be used.
                gpu_id = [1]

            self.__trainer = Trainer(
                callbacks=[RichModelSummary(max_depth=1), progress_bar],
                max_epochs=self.epochs,
                accelerator=self.config.participant["device_args"]["accelerator"],
                devices=gpu_id,
                logger=self.logger,
                log_every_n_steps=50,
                enable_checkpointing=False,
                enable_model_summary=False,
                enable_progress_bar=True,
                # deterministic=True
            )
        else:
            # NEBULA uses only CPU to distribute the nodes
            self.__trainer = Trainer(
                callbacks=[RichModelSummary(max_depth=1), progress_bar],
                max_epochs=self.epochs,
                accelerator=self.config.participant["device_args"]["accelerator"],
                devices="auto",
                logger=self.logger,
                log_every_n_steps=50,
                enable_checkpointing=False,
                enable_model_summary=False,
                enable_progress_bar=True,
                # deterministic=True
            )

    def get_global_model_parameters(self):
        return self.model.get_global_model_parameters()

    def set_parameter_second_aggregation(self, params):
        try:
            logging.info("Setting parameters in second aggregation...")
            self.model.load_state_dict(params)
        except:
            raise Exception("Error setting parameters")

    def get_model_parameters(self, bytes=False):
        if bytes:
            return self.serialize_model(self.model.state_dict())
        else:
            return self.model.state_dict()

    def get_hash_model(self):
        """
        Returns:
            str: SHA256 hash of model parameters
        """
        return hashlib.sha256(self.serialize_model()).hexdigest()

    def set_epochs(self, epochs):
        self.epochs = epochs

    ####
    # Model parameters serialization/deserialization
    # From https://pytorch.org/docs/stable/notes/serialization.html
    ####
    def serialize_model(self, model):
        try:
            buffer = io.BytesIO()
            # with gzip.GzipFile(fileobj=buffer, mode='wb') as f:
            #    torch.save(params, f)
            torch.save(model, buffer)
            return buffer.getvalue()
        except:
            raise Exception("Error serializing model")

    def deserialize_model(self, data):
        try:
            buffer = io.BytesIO(data)
            # with gzip.GzipFile(fileobj=buffer, mode='rb') as f:
            #    params_dict = torch.load(f, map_location='cpu')
            params_dict = torch.load(buffer, map_location="cpu")
            return OrderedDict(params_dict)
        except:
            raise Exception("Error decoding parameters")

    def set_model_parameters(self, params, initialize=False):
        try:
            if initialize:
                self.model.load_state_dict(params)
                self.model.global_load_state_dict(params)
                self.model.historical_load_state_dict(params)
            else:
                # First aggregation
                self.model.global_load_state_dict(params)
        except:
            raise Exception("Error setting parameters")

    def train(self):
        try:
            self.create_trainer()
            # torch.autograd.set_detect_anomaly(True)
            # TODO: It is necessary to train only the local model, save the history of the previous model and then load it, the global model is the aggregation of all the models.
            self.__trainer.fit(self.model, self.data)
            # Save local model as historical model (previous round)
            # It will be compared the next round during training local model (constrantive loss)
            # When aggregation in global model (first) and aggregation with similarities and weights (second), the historical model keeps inmutable
            logging.info("Saving historical model...")
            self.model.save_historical_model()
        except Exception as e:
            logging.exception(f"Error training model: {e}")
            logging.exception(traceback.format_exc())

    def test(self):
        try:
            self.create_trainer()
            self.__trainer.test(self.model, self.data, verbose=True)
        except Exception as e:
            logging.exception(f"Error testing model: {e}")
            logging.exception(traceback.format_exc())

    def get_model_weight(self):
        return (
            len(self.data.train_dataloader().dataset),
            len(self.data.test_dataloader().dataset),
        )

    def finalize_round(self):
        self.logger.global_step = self.logger.global_step + self.logger.local_step
        self.logger.local_step = 0
        self.round += 1
        self.logger.log_data({"Round": self.round}, step=self.logger.global_step)
        pass



import matplotlib
import matplotlib.pyplot as plt

matplotlib.use("Agg")
plt.switch_backend("Agg")
import torch
from torch import nn

from nebula.core.models.nebulamodel import NebulaModel


class SimpleMobileNetV1(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, 3, 32, 32)
        self.learning_rate = learning_rate
        self.criterion = torch.torch.nn.CrossEntropyLoss()

        def conv_dw(input_channels, num_classes, stride):
            return nn.Sequential(
                nn.Conv2d(
                    input_channels,
                    input_channels,
                    3,
                    stride,
                    1,
                    groups=input_channels,
                    bias=False,
                ),
                nn.BatchNorm2d(input_channels),
                nn.ReLU(inplace=True),
                nn.Conv2d(input_channels, num_classes, 1, 1, 0, bias=False),
                nn.BatchNorm2d(num_classes),
                nn.ReLU(inplace=True),
            )

        self.model = nn.Sequential(
            nn.Conv2d(3, 32, 3, 1, 1, bias=False),
            nn.BatchNorm2d(32),
            nn.ReLU(inplace=True),
            conv_dw(32, 64, 1),
            conv_dw(64, 128, 2),
            conv_dw(128, 128, 1),
            conv_dw(128, 256, 2),
            conv_dw(256, 256, 1),
            nn.AdaptiveAvgPool2d(1),
        )
        self.fc = nn.Linear(256, num_classes)

    def forward(self, x):
        x = self.model(x)
        x = x.view(-1, 256)
        x = self.fc(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



import numpy as np
from sklearn.svm import LinearSVC

from nebula.core.aggregation.aggregator import Aggregator


class FedAvgSVM(Aggregator):
    """
    Aggregator: Federated Averaging (FedAvg)
    Authors: McMahan et al.
    Year: 2016
    Note: This is a modified version of FedAvg for SVMs.
    """

    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)

    def run_aggregation(self, models):
        super().run_aggregation(models)

        models = list(models.values())

        total_samples = sum([y for _, y in models])

        coeff_accum = np.zeros_like(models[-1][0].coef_)
        intercept_accum = 0.0

        for model, w in models:
            if not isinstance(model, LinearSVC):
                return None
            coeff_accum += model.coef_ * w
            intercept_accum += model.intercept_ * w

        coeff_accum /= total_samples
        intercept_accum /= total_samples

        aggregated_svm = LinearSVC()
        aggregated_svm.coef_ = coeff_accum
        aggregated_svm.intercept_ = intercept_accum

        return aggregated_svm



import copy
import random

import numpy as np
import torch
from skimage.util import random_noise


def datapoison(
    dataset,
    indices,
    poisoned_persent,
    poisoned_ratio,
    targeted=False,
    target_label=3,
    noise_type="salt",
):
    """
    Function to add random noise of various types to the dataset.
    """
    new_dataset = copy.deepcopy(dataset)
    train_data = new_dataset.data
    targets = new_dataset.targets
    num_indices = len(indices)
    if type(noise_type) != str:
        noise_type = noise_type[0]

    if targeted == False:
        num_poisoned = int(poisoned_persent * num_indices)
        if num_indices == 0:
            return new_dataset
        if num_poisoned > num_indices:
            return new_dataset
        poisoned_indice = random.sample(indices, num_poisoned)

        for i in poisoned_indice:
            t = train_data[i]
            if noise_type == "salt":
                # Replaces random pixels with 1.
                poisoned = torch.tensor(random_noise(t, mode=noise_type, amount=poisoned_ratio))
            elif noise_type == "gaussian":
                # Gaussian-distributed additive noise.
                poisoned = torch.tensor(random_noise(t, mode=noise_type, mean=0, var=poisoned_ratio, clip=True))
            elif noise_type == "s&p":
                # Replaces random pixels with either 1 or low_val, where low_val is 0 for unsigned images or -1 for signed images.
                poisoned = torch.tensor(random_noise(t, mode=noise_type, amount=poisoned_ratio))
            elif noise_type == "nlp_rawdata":
                # for NLP data, change the word vector to 0 with p=poisoned_ratio
                poisoned = poison_to_nlp_rawdata(t, poisoned_ratio)
            else:
                print("ERROR: poison attack type not supported.")
                poisoned = t
            train_data[i] = poisoned
    else:
        for i in indices:
            if int(targets[i]) == int(target_label):
                t = train_data[i]
                poisoned = add_x_to_image(t)
                train_data[i] = poisoned
    new_dataset.data = train_data
    return new_dataset


def add_x_to_image(img):
    """
    Add a 10*10 pixels X at the top-left of an image
    """
    for i in range(0, 10):
        for j in range(0, 10):
            if i + j <= 9 or i == j:
                img[i][j] = 255
    return torch.tensor(img)


def poison_to_nlp_rawdata(text_data, poisoned_ratio):
    """
    for NLP data, change the word vector to 0 with p=poisoned_ratio
    """
    non_zero_vector_indice = [i for i in range(0, len(text_data)) if text_data[i][0] != 0]
    non_zero_vector_len = len(non_zero_vector_indice)

    num_poisoned_token = int(poisoned_ratio * non_zero_vector_len)
    if num_poisoned_token == 0:
        return text_data
    if num_poisoned_token > non_zero_vector_len:
        return text_data

    poisoned_token_indice = random.sample(non_zero_vector_indice, num_poisoned_token)
    zero_vector = torch.Tensor(np.zeros(len(text_data[0][0])))
    for i in poisoned_token_indice:
        text_data[i] = zero_vector
    return text_data



import glob
import json
import logging
import os

import numpy as np
import torch
from torch.utils.data import Dataset
from torchvision import transforms

from nebula.config.config import TRAINING_LOGGER
from nebula.core.datasets.nebuladataset import NebulaDataset

logging_training = logging.getLogger(TRAINING_LOGGER)


class RandomCrop:
    def __init__(self, size):
        if isinstance(size, int):
            self.size = (size, size)
        else:
            assert len(size) == 2
            self.size = size

    def __call__(self, sample):
        _input = sample

        if len(_input.shape) < 3:
            _input = np.expand_dims(_input, axis=2)

        h, w, _ = _input.shape
        oh, ow = self.size

        dh = h - oh
        dw = w - ow
        y = np.random.randint(0, dh) if dh > 0 else 0
        x = np.random.randint(0, dw) if dw > 0 else 0
        oh = oh if dh > 0 else h
        ow = ow if dw > 0 else w

        return _input[y : y + oh, x : x + ow, :]


class CenterCrop:
    def __init__(self, size):
        if isinstance(size, int):
            self.size = (size, size)
        else:
            assert len(size) == 2
            self.size = size

    def __call__(self, sample):
        _input = sample

        if len(_input.shape) < 3:
            _input = np.expand_dims(_input, axis=2)

        h, w, _ = _input.shape
        oh, ow = self.size
        y = (h - oh) // 2
        x = (w - ow) // 2

        return _input[y : y + oh, x : x + ow, :]


class MilitarySAR(Dataset):
    def __init__(self, name="soc", is_train=False, transform=None):
        self.is_train = is_train
        self.name = name

        self.data = []
        self.targets = []
        self.serial_numbers = []

        # Path to data is "data" folder in the same directory as this file
        self.path_to_data = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

        self.transform = transform

        # self._load_data(self.path_to_data)

        mode = "train" if self.is_train else "test"
        self.image_list = glob.glob(os.path.join(self.path_to_data, f"{self.name}/{mode}/*/*.npy"))
        self.label_list = glob.glob(os.path.join(self.path_to_data, f"{self.name}/{mode}/*/*.json"))
        self.image_list = sorted(self.image_list, key=os.path.basename)
        self.label_list = sorted(self.label_list, key=os.path.basename)
        assert len(self.image_list) == len(self.label_list)

    def __len__(self):
        return len(self.targets)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()

        _image = np.load(self.image_list[idx])
        with open(self.label_list[idx], encoding="utf-8") as f:
            label_info = json.load(f)
        _label = label_info["class_id"]
        # serial_number = label_info['serial_number']

        if self.transform:
            _image = self.transform(_image)

        return _image, _label

    def _load_metadata(self):
        self.targets = []
        self.serial_numbers = []
        for label_path in self.label_list:
            with open(label_path, encoding="utf-8") as f:
                label_info = json.load(f)
            self.targets.append(label_info["class_id"])
            self.serial_numbers.append(label_info["serial_number"])

    def get_targets(self):
        if not self.targets:
            logging_training.info(f"Loading Metadata for {self.__class__.__name__}")
            self._load_metadata()
        return self.targets

    # def _load_data(self, path):
    #     logging_training.info(f'Loading {self.__class__.__name__} dataset: {self.name} | is_train: {self.is_train} | from {self.path_to_data}')
    #     mode = 'train' if self.is_train else 'test'

    #     image_list = glob.glob(os.path.join(self.path_to_data, f'{self.name}/{mode}/*/*.npy'))
    #     label_list = glob.glob(os.path.join(self.path_to_data, f'{self.name}/{mode}/*/*.json'))
    #     image_list = sorted(image_list, key=os.path.basename)
    #     label_list = sorted(label_list, key=os.path.basename)

    #     for image_path, label_path in zip(image_list, label_list):
    #         self.data.append(np.load(image_path))

    #         with open(label_path, mode='r', encoding='utf-8') as f:
    #             _label = json.load(f)

    #         self.targets.append(_label['class_id'])
    #         self.serial_number.append(_label['serial_number'])

    #     self.data = np.array(self.data)
    #     self.targets = np.array(self.targets)


class MilitarySARDataset(NebulaDataset):
    def __init__(
        self,
        num_classes=10,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__(
            num_classes=num_classes,
            partition_id=partition_id,
            partitions_number=partitions_number,
            batch_size=batch_size,
            num_workers=num_workers,
            iid=iid,
            partition=partition,
            partition_parameter=partition_parameter,
            seed=seed,
            config=config,
        )

    def initialize_dataset(self):
        if self.train_set is None:
            self.train_set = self.load_militarysar_dataset(train=True)
        if self.test_set is None:
            self.test_set = self.load_militarysar_dataset(train=False)

        train_targets = self.train_set.get_targets()
        test_targets = self.test_set.get_targets()

        self.test_indices_map = list(range(len(self.test_set)))

        # Depending on the iid flag, generate a non-iid or iid map of the train set
        if self.iid:
            logging_training.info("Generating IID partition - Train")
            self.train_indices_map = self.generate_iid_map(self.train_set, self.partition, self.partition_parameter)
            logging_training.info("Generating IID partition - Test")
            self.local_test_indices_map = self.generate_iid_map(self.test_set, self.partition, self.partition_parameter)
        else:
            logging_training.info("Generating Non-IID partition - Train")
            self.train_indices_map = self.generate_non_iid_map(self.train_set, self.partition, self.partition_parameter)
            logging_training.info("Generating Non-IID partition - Test")
            self.local_test_indices_map = self.generate_non_iid_map(
                self.test_set, self.partition, self.partition_parameter
            )

        print(f"Length of train indices map: {len(self.train_indices_map)}")
        print(f"Lenght of test indices map (global): {len(self.test_indices_map)}")
        print(f"Length of test indices map (local): {len(self.local_test_indices_map)}")

    def load_militarysar_dataset(self, train=True):
        apply_transforms = [CenterCrop(88), transforms.ToTensor()]
        if train:
            apply_transforms = [RandomCrop(88), transforms.ToTensor()]

        return MilitarySAR(name="soc", is_train=train, transform=transforms.Compose(apply_transforms))

    def generate_non_iid_map(self, dataset, partition="dirichlet", partition_parameter=0.5):
        if partition == "dirichlet":
            partitions_map = self.dirichlet_partition(dataset, alpha=partition_parameter)
        elif partition == "percent":
            partitions_map = self.percentage_partition(dataset, percentage=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for Non-IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]

    def generate_iid_map(self, dataset, partition="balancediid", partition_parameter=2):
        if partition == "balancediid":
            partitions_map = self.balanced_iid_partition(dataset)
        elif partition == "unbalancediid":
            partitions_map = self.unbalanced_iid_partition(dataset, imbalance_factor=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]



import torch

from nebula.core.models.nebulamodel import NebulaModel


class SyscallModelAutoencoder(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.rand(1, input_channels)
        self.learning_rate = learning_rate

        self.criterion = torch.nn.MSELoss()

        self.fc1 = torch.nn.Linear(input_channels, 64)
        self.fc2 = torch.nn.Linear(64, 16)
        self.fc3 = torch.nn.Linear(16, 8)
        self.fc4 = torch.nn.Linear(8, 16)
        self.fc5 = torch.nn.Linear(16, 64)
        self.fc6 = torch.nn.Linear(64, input_channels)

        self.epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

    def encode(self, x):
        z = torch.relu(self.fc1(x))
        z = torch.relu(self.fc2(z))
        z = torch.relu(self.fc3(z))
        return z

    def decode(self, x):
        z = torch.relu(self.fc4(x))
        z = torch.relu(self.fc5(z))
        z = torch.relu(self.fc6(z))
        return z

    def forward(self, x):
        z = self.encode(x)
        z = self.decode(z)
        return z

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



import json
import os.path
import random
import shutil
import textwrap
from datetime import datetime

from eth_keys import keys
from web3 import Web3

w3 = Web3()


class BlockchainDeployer:
    """
    Creates files (docker-compose.yaml and genesis.json) for deploying blockchain network
    """

    def __init__(self, n_validator=3, config_dir=".", input_dir="."):
        # root dir of blockchain folder
        self.__input_dir = input_dir

        # config folder for storing generated files for deployment
        self.__config_dir = config_dir

        # random but static id of boot node to be assigned to all other nodes
        self.__boot_id = None

        # ip address of boot node (needs to be static)
        self.__boot_ip = "172.25.0.101"

        # ip address of non-validator node (needs to be static)
        self.__rpc_ip = "172.25.0.104"

        # ip address of oracle (needs to be static)
        self.__oracle_ip = "172.25.0.105"

        # temporary yaml parameter to store config before dump
        self.__yaml = ""

        # list of reserved addresses which need to be excluded in random address generation
        self.__reserved_addresses = set()

        # load original genesis dict
        self.__genesis = self.__load_genesis()

        # create blockchain directory in scenario's config directory
        self.__setup_dir()

        # add a boot node to the yaml file
        self.__add_boot_node()

        # add n validator nodes to the genesis.json and yaml file
        self.__add_validator(n_validator)

        # add non-validator node to the yaml file
        self.__add_rpc()

        # add oracle node to the genesis.json and yaml file
        self.__add_oracle()

        # dump config files into scenario's config directory
        self.__export_config()

    def __setup_dir(self) -> None:
        if not os.path.exists(self.__config_dir):
            os.makedirs(self.__config_dir, exist_ok=True)

    def __get_unreserved_address(self) -> tuple[int, int]:
        """
        Computes a randomized port and last 8 bits of an ip address, where both are not yet used
        Returns: Randomized and unreserved lat 8 bit of ip and port

        """

        # extract reserved ports and ip addresses
        reserved_ips = [address[0] for address in self.__reserved_addresses]
        reserved_ports = [address[1] for address in self.__reserved_addresses]

        # get randomized ip and port in range still unreserved
        ip = random.choice([number for number in range(10, 254) if number not in reserved_ips])
        port = random.choice([number for number in range(30310, 30360) if number not in reserved_ports])

        # add network address to list of reserved addresses
        self.__reserved_addresses.add((ip, port))
        return ip, port

    def __copy_dir(self, source_path) -> None:
        """
        Copy blockchain folder with current files such as chaincode to config folder
        Args:
            source_path: Path of dir to copy

        Returns: None

        """

        curr_path = os.path.dirname(os.path.abspath(__file__))

        if not os.path.exists(self.__config_dir):
            os.makedirs(self.__config_dir, exist_ok=True)

        target_dir = os.path.join(self.__config_dir, source_path)
        source_dir = os.path.join(curr_path, source_path)
        shutil.copytree(str(source_dir), target_dir, dirs_exist_ok=True)

    @staticmethod
    def __load_genesis() -> dict[str, int | str | dict]:
        """
        Load original genesis config
        Returns: Genesis json dict

        """
        return {
            "config": {
                "chainId": 19265019,  # unique id not used by any public Ethereum network
                # block number at which the defined EIP hard fork policies are applied
                "homesteadBlock": 0,
                "eip150Block": 0,
                "eip155Block": 0,
                "eip158Block": 0,
                "byzantiumBlock": 0,
                "constantinopleBlock": 0,
                "petersburgBlock": 0,
                "istanbulBlock": 0,
                "muirGlacierBlock": 0,
                "berlinBlock": 0,
                # Proof-of-Authority settings
                "clique": {
                    "period": 1,
                    "epoch": 10000,
                },  # block time (time in seconds between two blocks)  # number of blocks after reset the pending votes
            },
            # unique continuous id of transactions used by PoA
            "nonce": "0x0",
            # UNIX timestamp of block creation
            "timestamp": "0x5a8efd25",
            # strictly formated string containing all public wallet addresses of all validators (PoA)
            # will be replaced by public addresses of randomly generated validator node
            "extraData": "0x0000000000000000000000000000000000000000000000000000000000000000187c1c14c75bA185A59c621Fbe5dda26D488852DF20C144e8aE3e1aCF7071C4883B759D1B428e7930000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000",
            # maximum gas (computational cost) per transaction
            "gasLimit": "9000000000000",  # "8000000" is default for Ethereum but too low for heavy load
            # difficulty for PoW
            "difficulty": "0x1",
            # root hash of block
            "mixHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
            # validator of genesis block
            "coinbase": "0x0000000000000000000000000000000000000000",
            # prefunded public wallet addresses (Oracle)
            "alloc": {
                # will be replaced by Oracle's randomized address
                "0x61DE01FcD560da4D6e05E58bCD34C8Dc92CE36D1": {
                    "balance": "0x200000000000000000000000000000000000000000000000000000000000000"
                }
            },
            # block number of genesis block
            "number": "0x0",
            # gas used to validate genesis block
            "gasUsed": "0x0",
            # hash of parent block (0x0 since first block)
            "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
        }

    def __add_boot_node(self) -> None:
        """
        Adds boot node to docker-compose.yaml
        Returns: None

        """

        # create random private key and create account from it
        acc = w3.eth.account.create()

        # store id of boot node to be inserted into all other nodes
        self.__boot_id = str(keys.PrivateKey(acc.key).public_key)[2:]

        # add service to yaml string
        self.__yaml += textwrap.dedent(
            f"""
            geth-bootnode:
                hostname: geth-bootnode
                environment:
                  - nodekeyhex={w3.to_hex(acc.key)[2:]}
                build:
                  dockerfile: {self.__input_dir}/geth/boot.dockerfile
                container_name: boot
                networks:
                  chainnet:
                    ipv4_address: {self.__boot_ip}
            """
        )

    def __add_validator(self, cnt) -> None:
        """
        Randomly generates and adds number(cnt) of validator nodes to yaml and genesis.json
        Args:
            cnt: number of validator nodes to cresate

        Returns: None

        """
        validator_addresses = list()

        for id in range(cnt):
            # create random private key and create account from it
            acc = w3.eth.account.create()
            validator_addresses.append(acc.address[2:])

            # get random network address
            ip, port = self.__get_unreserved_address()

            self.__yaml += textwrap.dedent(
                f"""
                geth-validator-{id}:
                    hostname: geth-validator-{id}
                    depends_on:
                      - geth-bootnode
                    environment:
                      - address={acc.address}
                      - bootnodeId={self.__boot_id}
                      - bootnodeIp={self.__boot_ip}
                      - port={port}
                    build:
                      dockerfile: {self.__input_dir}/geth/validator.dockerfile
                      args:
                        privatekey: {w3.to_hex(acc.key)[2:]}
                        password: {w3.to_hex(w3.eth.account.create().key)}
                    container_name: validator_{id}
                    networks:
                      chainnet:
                        ipv4_address: 172.25.0.{ip}
                """
            )

        # create specific Ethereum extra data string for PoA with all public addresses of validators
        extra_data = "0x" + "0" * 64 + "".join([a for a in validator_addresses]) + 65 * "0" + 65 * "0"
        self.__genesis["extraData"] = extra_data

    def __add_oracle(self) -> None:
        """
        Adds Oracle node to yaml and genesis.json
        Returns: None

        """

        # create random private key and create account from it
        acc = w3.eth.account.create()

        # prefund oracle by allocating all funds to its public wallet address
        self.__genesis["alloc"] = {
            acc.address: {"balance": "0x200000000000000000000000000000000000000000000000000000000000000"}
        }

        self.__yaml += textwrap.dedent(
            f"""
            oracle:
               hostname: oracle
               depends_on:
                 - geth-rpc
                 - geth-bootnode
               environment:
                 - PRIVATE_KEY={w3.to_hex(acc.key)[2:]}
                 - RPC_IP={self.__rpc_ip}
               build:
                 dockerfile: {self.__input_dir}/geth/oracle.dockerfile
                 context: {self.__input_dir}
               ports:
                 - 8081:8081
               container_name: oracle
               networks:
                 chainnet:
                   ipv4_address: {self.__oracle_ip}
            """
        )

    def __add_rpc(self):
        """
        Add non-validator node to yaml
        Returns: None

        """
        # create random private key and create account from it
        acc = w3.eth.account.create()

        self.__yaml += textwrap.dedent(
            f"""
            geth-rpc:
                 hostname: geth-rpc
                 depends_on:
                   - geth-bootnode
                 environment:
                   - address={acc.address}
                   - bootnodeId={self.__boot_id}
                   - bootnodeIp={self.__boot_ip}
                 build:
                   dockerfile: {self.__input_dir}/geth/rpc.dockerfile
                 ports:
                   - 8545:8545
                 container_name: rpc
                 networks:
                   chainnet:
                     ipv4_address: {self.__rpc_ip}
            """
        )

    def __add_network(self) -> None:
        """
        Adds network config to docker-compose.yaml to create a private network for docker compose
        Returns: None

        """
        self.__yaml += textwrap.dedent(
            """
            networks:
              chainnet:
                name: chainnet
                driver: bridge
                ipam:
                  config:
                  - subnet: 172.25.0.0/24
            """
        )

    def __export_config(self) -> None:
        """
        Writes configured yaml and genesis files to config folder for deplyoment
        Returns: None

        """

        # format yaml and add docker compose properties
        final_str = textwrap.indent(f"""{self.__yaml}""", "  ")

        self.__yaml = textwrap.dedent(
            """
                    version: "3.8"
                    name: blockchain
                    services:
                    """
        )

        self.__yaml += final_str

        # add network config last
        self.__add_network()

        with open(f"{self.__config_dir}/blockchain-docker-compose.yml", "w+") as file:
            file.write(self.__yaml)

        with open(f"{self.__input_dir}/geth/genesis.json", "w+") as file:
            json.dump(self.__genesis, file, indent=4)

        source = os.path.join(self.__input_dir, "geth", "genesis.json")
        shutil.copy(source, os.path.join(self.__config_dir, "genesis.json"))

        source = os.path.join(self.__input_dir, "chaincode", "reputation_system.sol")
        shutil.copy(source, os.path.join(self.__config_dir, "reputation_system.sol"))


if __name__ == "__main__":
    b = BlockchainDeployer(
        n_validator=3,
        config_dir=os.path.join("deployments", datetime.now().strftime("%Y-%m-%d_%H-%M")),
    )



import json
import logging
import math
import os
import pickle
from os.path import exists

import pandas as pd
from hashids import Hashids
from scipy.stats import entropy

from nebula.addons.trustworthiness import calculation

hashids = Hashids()
logger = logging.getLogger(__name__)
dirname = os.path.dirname(__file__)


def count_class_samples(scenario_name, dataloaders_files):
    """
    Counts the number of samples by class.

    Args:
        scenario_name (string): Name of the scenario.
        dataloaders_files (list): Files that contain the dataloaders.

    """

    result = {}
    dataloaders = []

    for file in dataloaders_files:
        with open(file, "rb") as f:
            dataloader = pickle.load(f)
            dataloaders.append(dataloader)

    for dataloader in dataloaders:
        for batch, labels in dataloader:
            for b, label in zip(batch, labels, strict=False):
                l = hashids.encode(label.item())
                if l in result:
                    result[l] += 1
                else:
                    result[l] = 1

    name_file = f"{dirname}/files/{scenario_name}/count_class.json"
    with open(name_file, "w") as f:
        json.dump(result, f)


def get_entropy(client_id, scenario_name, dataloader):
    """
    Get the entropy of each client in the scenario.

    Args:
        client_id (int): The client id.
        scenario_name (string): Name of the scenario.
        dataloaders_files (list): Files that contain the dataloaders.

    """
    result = {}
    client_entropy = {}

    name_file = f"{dirname}/files/{scenario_name}/entropy.json"
    if os.path.exists(name_file):
        with open(name_file) as f:
            client_entropy = json.load(f)

    client_id_hash = hashids.encode(client_id)

    for batch, labels in dataloader:
        for b, label in zip(batch, labels, strict=False):
            l = hashids.encode(label.item())
            if l in result:
                result[l] += 1
            else:
                result[l] = 1

    n = len(dataloader)
    entropy_value = entropy([x / n for x in result.values()], base=2)
    client_entropy[client_id_hash] = entropy_value
    with open(name_file, "w") as f:
        json.dump(client_entropy, f)


def read_csv(filename):
    """
    Read a CSV file.

    Args:
        filename (string): Name of the file.

    Returns:
        object: The CSV readed.

    """
    if exists(filename):
        return pd.read_csv(filename)


def check_field_filled(factsheet_dict, factsheet_path, value, empty=""):
    """
    Check if the field in the factsheet file is filled or not.

    Args:
        factsheet_dict (dict): The factshett dict.
        factsheet_path (list): The factsheet field to check.
        value (float): The value to add in the field.
        empty (string): If the value could not be appended, the empty string is returned.

    Returns:
        float: The value added in the factsheet or empty if the value could not be appened

    """
    if factsheet_dict[factsheet_path[0]][factsheet_path[1]]:
        return factsheet_dict[factsheet_path[0]][factsheet_path[1]]
    elif value != "" and value != "nan":
        if type(value) != str and type(value) != list:
            if math.isnan(value):
                return 0
            else:
                return value
        else:
            return value
    else:
        return empty


def get_input_value(input_docs, inputs, operation):
    """
    Gets the input value from input document and apply the metric operation on the value.

    Args:
        inputs_docs (map): The input document map.
        inputs (list): All the inputs.
        operation (string): The metric operation.

    Returns:
        float: The metric value

    """

    input_value = None
    args = []
    for i in inputs:
        source = i.get("source", "")
        field = i.get("field_path", "")
        input_doc = input_docs.get(source, None)
        if input_doc is None:
            logger.warning(f"{source} is null")
        else:
            input = get_value_from_path(input_doc, field)
            args.append(input)
    try:
        operationFn = getattr(calculation, operation)
        input_value = operationFn(*args)
    except TypeError:
        logger.warning(f"{operation} is not valid")

    return input_value


def get_value_from_path(input_doc, path):
    """
    Gets the input value from input document by path.

    Args:
        inputs_doc (map): The input document map.
        path (string): The field name of the input value of interest.

    Returns:
        float: The input value from the input document

    """

    d = input_doc
    for nested_key in path.split("/"):
        temp = d.get(nested_key)
        if isinstance(temp, dict):
            d = d.get(nested_key)
        else:
            return temp
    return None


def write_results_json(out_file, dict):
    """
    Writes the result to JSON.

    Args:
        out_file (string): The output file.
        dict (dict): The object to be witten into JSON.

    Returns:
        float: The input value from the input document

    """

    with open(out_file, "a") as f:
        json.dump(dict, f, indent=4)


def save_results_csv(
    scenario_name: str,
    id: int,
    bytes_sent: int,
    bytes_recv: int,
    accuracy: float,
    loss: float,
    finish: bool,
):
    outdir = f"{dirname}/files/{scenario_name}"
    filename = "data_results.csv"
    data_results_file = os.path.join(outdir, filename)
    if exists(data_results_file):
        df = pd.read_csv(data_results_file)
    else:
        # Crear un DataFrame con columnas especificadas
        df = pd.DataFrame(columns=["id", "bytes_sent", "bytes_recv", "accuracy", "loss", "finish"])

    try:
        if id not in df["id"].values:
            # Si no existe, agregar una nueva entrada con el ID del nodo
            df = pd.concat(
                [
                    df,
                    pd.DataFrame({
                        "id": id,
                        "bytes_sent": None,
                        "bytes_recv": None,
                        "accuracy": None,
                        "loss": None,
                        "finish": False,
                    }),
                ],
                ignore_index=True,
            )

            df.to_csv(data_results_file, encoding="utf-8", index=False)
        else:
            if bytes_sent is not None:
                df.loc[df["id"] == id, "bytes_sent"] = bytes_sent
            if bytes_recv is not None:
                df.loc[df["id"] == id, "bytes_recv"] = bytes_recv
            if accuracy is not None:
                df.loc[df["id"] == id, "accuracy"] = accuracy
            if loss is not None:
                df.loc[df["id"] == id, "loss"] = loss
            if finish:
                df.loc[df["id"] == id, "finish"] = True
            df.to_csv(data_results_file, encoding="utf-8", index=False)

    except Exception as e:
        logger.warning(e)



import asyncio
import datetime
import sqlite3

import aiosqlite
from argon2 import PasswordHasher

user_db_file_location = "databases/users.db"
node_db_file_location = "databases/nodes.db"
scenario_db_file_location = "databases/scenarios.db"
notes_db_file_location = "databases/notes.db"

_node_lock = asyncio.Lock()

PRAGMA_SETTINGS = [
    "PRAGMA journal_mode=WAL;",
    "PRAGMA synchronous=NORMAL;",
    "PRAGMA journal_size_limit=1048576;",
    "PRAGMA cache_size=10000;",
    "PRAGMA temp_store=MEMORY;",
    "PRAGMA cache_spill=0;",
]


async def setup_database(db_file_location):
    async with aiosqlite.connect(db_file_location) as db:
        for pragma in PRAGMA_SETTINGS:
            await db.execute(pragma)
        await db.commit()


async def initialize_databases():
    await setup_database(user_db_file_location)
    await setup_database(node_db_file_location)
    await setup_database(scenario_db_file_location)
    await setup_database(notes_db_file_location)

    async with aiosqlite.connect(user_db_file_location) as conn:
        _c = await conn.cursor()
        await _c.execute(
            """
            CREATE TABLE IF NOT EXISTS users (
                user TEXT PRIMARY KEY,
                password TEXT NOT NULL,
                role TEXT NOT NULL
            );
            """
        )
        await conn.commit()

    async with aiosqlite.connect(node_db_file_location) as conn:
        _c = await conn.cursor()
        await _c.execute(
            """
            CREATE TABLE IF NOT EXISTS nodes (
                uid TEXT PRIMARY KEY,
                idx TEXT NOT NULL,
                ip TEXT NOT NULL,
                port TEXT NOT NULL,
                role TEXT NOT NULL,
                neighbors TEXT NOT NULL,
                latitude TEXT NOT NULL,
                longitude TEXT NOT NULL,
                timestamp TEXT NOT NULL,
                federation TEXT NOT NULL,
                round TEXT NOT NULL,
                scenario TEXT NOT NULL,
                hash TEXT NOT NULL
            );
            """
        )
        await conn.commit()

    async with aiosqlite.connect(scenario_db_file_location) as conn:
        _c = await conn.cursor()
        await _c.execute(
            """
            CREATE TABLE IF NOT EXISTS scenarios (
                name TEXT PRIMARY KEY,
                start_time TEXT NOT NULL,
                end_time TEXT NOT NULL,
                title TEXT NOT NULL,
                description TEXT NOT NULL,
                status TEXT NOT NULL,
                network_subnet TEXT NOT NULL,
                model TEXT NOT NULL,
                dataset TEXT NOT NULL,
                rounds TEXT NOT NULL,
                role TEXT NOT NULL
            );
            """
        )
        await conn.commit()

    async with aiosqlite.connect(notes_db_file_location) as conn:
        _c = await conn.cursor()
        await _c.execute(
            """
            CREATE TABLE IF NOT EXISTS notes (
                scenario TEXT PRIMARY KEY,
                scenario_notes TEXT NOT NULL
            );
            """
        )
        await conn.commit()


def list_users(all_info=False):
    with sqlite3.connect(user_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        c.execute("SELECT * FROM users")
        result = c.fetchall()

    if not all_info:
        result = [user["user"] for user in result]

    return result


def get_user_info(user):
    with sqlite3.connect(user_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()

        command = "SELECT * FROM users WHERE user = ?"
        c.execute(command, (user,))
        result = c.fetchone()

    return result


def verify(user, password):
    ph = PasswordHasher()
    with sqlite3.connect(user_db_file_location) as conn:
        c = conn.cursor()

        c.execute("SELECT password FROM users WHERE user = ?", (user,))
        result = c.fetchone()
        if result:
            try:
                return ph.verify(result[0], password)
            except:
                return False
    return False


def verify_hash_algorithm(user):
    user = user.upper()
    argon2_prefixes = ("$argon2i$", "$argon2id$")

    with sqlite3.connect(user_db_file_location) as conn:
        c = conn.cursor()

        c.execute("SELECT password FROM users WHERE user = ?", (user,))
        result = c.fetchone()
        if result:
            password_hash = result[0]
            return password_hash.startswith(argon2_prefixes)

    return False


def delete_user_from_db(user):
    with sqlite3.connect(user_db_file_location) as conn:
        c = conn.cursor()
        c.execute("DELETE FROM users WHERE user = ?", (user,))


def add_user(user, password, role):
    ph = PasswordHasher()
    with sqlite3.connect(user_db_file_location) as conn:
        c = conn.cursor()
        c.execute(
            "INSERT INTO users VALUES (?, ?, ?)",
            (user.upper(), ph.hash(password), role),
        )


def update_user(user, password, role):
    ph = PasswordHasher()
    with sqlite3.connect(user_db_file_location) as conn:
        c = conn.cursor()
        c.execute(
            "UPDATE users SET password = ?, role = ? WHERE user = ?",
            (ph.hash(password), role, user.upper()),
        )


def list_nodes(scenario_name=None, sort_by="idx"):
    # list all nodes in the database
    try:
        with sqlite3.connect(node_db_file_location) as conn:
            c = conn.cursor()

            if scenario_name:
                command = "SELECT * FROM nodes WHERE scenario = ? ORDER BY " + sort_by + ";"
                c.execute(command, (scenario_name,))
            else:
                command = "SELECT * FROM nodes ORDER BY " + sort_by + ";"
                c.execute(command)

            result = c.fetchall()

            return result
    except sqlite3.Error as e:
        print(f"Error occurred while listing nodes: {e}")
        return None


def list_nodes_by_scenario_name(scenario_name):
    try:
        with sqlite3.connect(node_db_file_location) as conn:
            c = conn.cursor()

            command = "SELECT * FROM nodes WHERE scenario = ? ORDER BY CAST(idx AS INTEGER) ASC;"
            c.execute(command, (scenario_name,))
            result = c.fetchall()

            return result
    except sqlite3.Error as e:
        print(f"Error occurred while listing nodes by scenario name: {e}")
        return None


async def update_node_record(
    node_uid,
    idx,
    ip,
    port,
    role,
    neighbors,
    latitude,
    longitude,
    timestamp,
    federation,
    federation_round,
    scenario,
    run_hash,
):
    # Check if the node record with node_uid and scenario already exists in the database
    # If it does, update the record
    # If it does not, create a new record
    # _conn = sqlite3.connect(node_db_file_location)
    global _node_lock
    async with _node_lock:
        async with aiosqlite.connect(node_db_file_location) as conn:
            _c = await conn.cursor()

            command = "SELECT * FROM nodes WHERE uid = ? AND scenario = ?;"
            await _c.execute(command, (node_uid, scenario))
            result = await _c.fetchone()

            if result is None:
                # Create a new record
                await _c.execute(
                    "INSERT INTO nodes VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        node_uid,
                        idx,
                        ip,
                        port,
                        role,
                        neighbors,
                        latitude,
                        longitude,
                        timestamp,
                        federation,
                        federation_round,
                        scenario,
                        run_hash,
                    ),
                )
            else:
                # Update the record
                command = "UPDATE nodes SET idx = ?, ip = ?, port = ?, role = ?, neighbors = ?, latitude = ?, longitude = ?, timestamp = ?, federation = ?, round = ?, hash = ? WHERE uid = ? AND scenario = ?;"
                await _c.execute(
                    command,
                    (
                        idx,
                        ip,
                        port,
                        role,
                        neighbors,
                        latitude,
                        longitude,
                        timestamp,
                        federation,
                        federation_round,
                        run_hash,
                        node_uid,
                        scenario,
                    ),
                )

            await conn.commit()


def remove_all_nodes():
    with sqlite3.connect(node_db_file_location) as conn:
        c = conn.cursor()
        command = "DELETE FROM nodes;"
        c.execute(command)


def remove_nodes_by_scenario_name(scenario_name):
    with sqlite3.connect(node_db_file_location) as conn:
        c = conn.cursor()
        command = "DELETE FROM nodes WHERE scenario = ?;"
        c.execute(command, (scenario_name,))


def get_run_hashes_scenario(scenario_name):
    with sqlite3.connect(node_db_file_location) as conn:
        c = conn.cursor()
        command = "SELECT DISTINCT idx, hash FROM nodes WHERE scenario = ?;"
        c.execute(command, (scenario_name,))
        result = c.fetchall()
        result_hashes = [(f"participant_{node[0]}", node[1]) for node in result]

        return result_hashes


def get_all_scenarios(sort_by="start_time"):
    with sqlite3.connect(scenario_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        if sort_by == "start_time":
            command = """
            SELECT * FROM scenarios
            ORDER BY strftime('%Y-%m-%d %H:%M:%S', substr(start_time, 7, 4) || '-' || substr(start_time, 4, 2) || '-' || substr(start_time, 1, 2) || ' ' || substr(start_time, 12, 8));
            """
            c.execute(command)
        else:
            command = "SELECT * FROM scenarios ORDER BY ?;"
            c.execute(command, (sort_by,))
        result = c.fetchall()

    return result


def get_all_scenarios_and_check_completed(sort_by="start_time"):
    with sqlite3.connect(scenario_db_file_location) as _conn:
        _conn.row_factory = sqlite3.Row
        _c = _conn.cursor()
        if sort_by == "start_time":
            command = """
            SELECT * FROM scenarios
            ORDER BY strftime('%Y-%m-%d %H:%M:%S', substr(start_time, 7, 4) || '-' || substr(start_time, 4, 2) || '-' || substr(start_time, 1, 2) || ' ' || substr(start_time, 12, 8));
            """
            _c.execute(command)
        else:
            command = "SELECT * FROM scenarios ORDER BY ?;"
            _c.execute(command, (sort_by,))
        _c.execute(command)
        result = _c.fetchall()

        for scenario in result:
            if scenario["status"] == "running":
                if check_scenario_federation_completed(scenario["name"]):
                    scenario_set_status_to_completed(scenario["name"])
                    result = get_all_scenarios()

    return result


def scenario_update_record(
    scenario_name,
    start_time,
    end_time,
    title,
    description,
    status,
    network_subnet,
    model,
    dataset,
    rounds,
    role,
):
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()

    command = "SELECT * FROM scenarios WHERE name = ?;"
    _c.execute(command, (scenario_name,))
    result = _c.fetchone()

    if result is None:
        # Create a new record
        _c.execute(
            "INSERT INTO scenarios VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            (
                scenario_name,
                start_time,
                end_time,
                title,
                description,
                status,
                network_subnet,
                model,
                dataset,
                rounds,
                role,
            ),
        )
    else:
        # Update the record
        command = "UPDATE scenarios SET start_time = ?, end_time = ?, title = ?, description = ?, status = ?, network_subnet = ?, model = ?, dataset = ?, rounds = ?, role = ? WHERE name = ?;"
        _c.execute(
            command,
            (
                start_time,
                end_time,
                title,
                description,
                status,
                network_subnet,
                model,
                dataset,
                rounds,
                role,
                scenario_name,
            ),
        )

    _conn.commit()
    _conn.close()


def scenario_set_all_status_to_finished():
    # Set all running scenarios to finished and update the end_time to the current time
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()

    command = "UPDATE scenarios SET status = 'finished', end_time = ? WHERE status = 'running';"
    current_time = str(datetime.datetime.now())
    _c.execute(command, (current_time,))

    _conn.commit()
    _conn.close()


def scenario_set_status_to_finished(scenario_name):
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()

    command = "UPDATE scenarios SET status = 'finished', end_time = ? WHERE name = ?;"
    current_time = str(datetime.datetime.now())
    _c.execute(command, (current_time, scenario_name))

    _conn.commit()
    _conn.close()


def scenario_set_status_to_completed(scenario_name):
    try:
        with sqlite3.connect(scenario_db_file_location) as _conn:
            _c = _conn.cursor()
            command = "UPDATE scenarios SET status = 'completed' WHERE name = ?;"
            _c.execute(command, (scenario_name,))
            _conn.commit()
            _conn.close()
    except sqlite3.Error as e:
        print(f"Database error: {e}")


def get_running_scenario():
    with sqlite3.connect(scenario_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        command = "SELECT * FROM scenarios WHERE status = ? OR status = ?;"
        c.execute(command, ("running", "completed"))
        result = c.fetchone()

    return result


def get_completed_scenario():
    with sqlite3.connect(scenario_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        command = "SELECT * FROM scenarios WHERE status = ?;"
        c.execute(command, ("completed",))
        result = c.fetchone()

    return result


def get_scenario_by_name(scenario_name):
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()
    command = "SELECT * FROM scenarios WHERE name = ?;"
    _c.execute(command, (scenario_name,))
    result = _c.fetchone()

    _conn.commit()
    _conn.close()

    return result


def remove_scenario_by_name(scenario_name):
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()

    command = "DELETE FROM scenarios WHERE name = ?;"
    _c.execute(command, (scenario_name,))

    _conn.commit()
    _conn.close()


def check_scenario_federation_completed(scenario_name):
    try:
        # Connect to the scenario database to get the total rounds for the scenario
        with sqlite3.connect(scenario_db_file_location) as conn:
            conn.row_factory = sqlite3.Row
            c = conn.cursor()
            c.execute("SELECT rounds FROM scenarios WHERE name = ?;", (scenario_name,))
            scenario = c.fetchone()

            if not scenario:
                raise ValueError(f"Scenario '{scenario_name}' not found.")

            total_rounds = scenario["rounds"]

        # Connect to the node database to check the rounds for each node
        with sqlite3.connect(node_db_file_location) as conn:
            conn.row_factory = sqlite3.Row
            c = conn.cursor()
            c.execute("SELECT round FROM nodes WHERE scenario = ?;", (scenario_name,))
            nodes = c.fetchall()

            if len(nodes) == 0:
                return False

            # Check if all nodes have completed the total rounds
            return all(node["round"] == total_rounds for node in nodes)

    except sqlite3.Error as e:
        print(f"Database error: {e}")
        return False
    except Exception as e:
        print(f"An error occurred: {e}")
        return False


def check_scenario_with_role(role, scenario_name):
    _conn = sqlite3.connect(scenario_db_file_location)
    _c = _conn.cursor()
    command = "SELECT * FROM scenarios WHERE role = ? AND name = ?;"
    _c.execute(command, (role, scenario_name))
    result = _c.fetchone()

    _conn.commit()
    _conn.close()

    return result


def save_notes(scenario, notes):
    try:
        with sqlite3.connect(notes_db_file_location) as conn:
            c = conn.cursor()
            c.execute(
                """
                INSERT INTO notes (scenario, scenario_notes) VALUES (?, ?)
                ON CONFLICT(scenario) DO UPDATE SET scenario_notes = excluded.scenario_notes;
                """,
                (scenario, notes),
            )
            conn.commit()
    except sqlite3.IntegrityError as e:
        print(f"SQLite integrity error: {e}")
    except sqlite3.Error as e:
        print(f"SQLite error: {e}")


def get_notes(scenario):
    with sqlite3.connect(notes_db_file_location) as conn:
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        c.execute("SELECT * FROM notes WHERE scenario = ?;", (scenario,))
        result = c.fetchone()

    return result


def remove_note(scenario):
    with sqlite3.connect(notes_db_file_location) as conn:
        c = conn.cursor()
        c.execute("DELETE FROM notes WHERE scenario = ?;", (scenario,))
        conn.commit()


if __name__ == "__main__":
    print(list_users())



import gc

import torch

from nebula.core.aggregation.aggregator import Aggregator


class FedAvg(Aggregator):
    """
    Aggregator: Federated Averaging (FedAvg)
    Authors: McMahan et al.
    Year: 2016
    """

    def __init__(self, config=None, **kwargs):
        super().__init__(config, **kwargs)

    def run_aggregation(self, models):
        super().run_aggregation(models)

        models = list(models.values())

        total_samples = float(sum(weight for _, weight in models))

        if total_samples == 0:
            raise ValueError("Total number of samples must be greater than zero.")

        last_model_params = models[-1][0]
        accum = {layer: torch.zeros_like(param, dtype=torch.float32) for layer, param in last_model_params.items()}

        with torch.no_grad():
            for model_parameters, weight in models:
                normalized_weight = weight / total_samples
                for layer in accum:
                    accum[layer].add_(
                        model_parameters[layer].to(accum[layer].dtype),
                        alpha=normalized_weight,
                    )

        del models
        gc.collect()

        # self.print_model_size(accum)
        return accum
